from django import forms, urls
from django.contrib import admin, auth, messages
from django.contrib.admin.views.main import ChangeList
from django.core import mail
from django.db import connection
from django.db.models import Count, F, Q, Max
from django.http import HttpResponseRedirect
from django.shortcuts import get_object_or_404
from django.template import Template
from django.template.response import TemplateResponse
from django.urls import path, reverse, NoReverseMatch
from django.utils import timezone
from django.utils.html import format_html
from django.utils.http import urlencode
from django.utils.safestring import mark_safe
from django_better_admin_arrayfield.admin.mixins import DynamicArrayMixin
from polymorphic.admin import (PolymorphicParentModelAdmin,
                               PolymorphicChildModelAdmin,
                               PolymorphicChildModelFilter)
import sys, importlib, time, types
from urllib.parse import unquote, parse_qsl

from .forms import ProgramAdminForm, FormAdminForm, StockBlockAdminForm, \
    CustomBlockAdminForm, CollectionBlockAdminForm, SubmissionAdminForm, \
    SubmissionItemAdminForm, EmailAdminForm, MoveBlocksAdminForm
from .models import Program, Form, FormLabel, FormBlock, FormDependency, \
    CustomBlock, CollectionBlock, SubmissionRecord
from .filetype import FileType
from .plugins import get_matching_plugin
from .signals import register_program_settings, register_form_settings
from .utils import send_email, submission_link


class FormativeAdminSite(admin.AdminSite):
    def __init__(self, *args, **kwargs):
        self.submissions_registered = None
        super().__init__(*args, **kwargs)
    
    def register_submission_models(self):
        if Form._meta.db_table in connection.introspection.table_names():
            for form in Form.objects.exclude(status=Form.Status.DRAFT):
                self.register(form.model, SubmissionAdmin)
                if form.item_model:
                    self.register(form.item_model, SubmissionItemAdmin)
        self.submissions_registered = True
    
    def catch_all_view(self, request, url):
        if not self.submissions_registered:
            self.get_app_list(request)
            return HttpResponseRedirect(request.get_full_path()) # better than 0
            # TODO a middleware is necessary to solve this?
        
        return super().catch_all_view(request, url)
    
    def get_app_list(self, request):
        if not self.submissions_registered: self.register_submission_models()
        
        # unlike normal Django, we might have had changes to the admin urls
        urls.clear_url_caches()
        if 'urls' in sys.modules: importlib.reload(sys.modules['urls'])
        
        return super().get_app_list(request)
    
    def app_index(self, request, *args, **kwargs):
        if not self.submissions_registered: self.get_app_list(request)
        
        return super().app_index(request, *args, **kwargs)


site = FormativeAdminSite()

site.register(auth.models.Group, auth.admin.GroupAdmin)
site.register(auth.models.User, auth.admin.UserAdmin)


@admin.register(Program, site=site)
class ProgramAdmin(admin.ModelAdmin, DynamicArrayMixin):
    list_display = ('name', 'created')
    form = ProgramAdminForm
    
    def get_fieldsets(self, request, obj=None):
        fields = self.get_fields(request, obj)
        
        main = (None, {'fields': fields})
        if not obj: return [main]
        
        responses = register_program_settings.send(self)
        admin_fields = { k: v for _, r in responses for k, v in r.items() }
        if not admin_fields: return [main]
        return [
            main,
            ('Options', {'fields': list(admin_fields)}),
        ]
    
    def get_readonly_fields(self, request, obj=None):
        fields = super().get_readonly_fields(request, obj)
        if obj: fields += ('slug',)
        return fields


class FormChangeList(ChangeList):
    def url_for_result(self, result):
        pk = getattr(result, self.pk_attname)
        url = reverse('admin:%s_formblock_formlist' % (self.opts.app_label,),
                       args=(int(pk),),
                       current_app=self.model_admin.admin_site.name)
        return url + '?page=1'


@admin.register(Form, site=site)
class FormAdmin(admin.ModelAdmin):
    list_display = ('name', 'program', 'created', 'modified')
    list_filter = ('program',)
    form = FormAdminForm
    
    def get_changelist(self, request, **kwargs):
        return FormChangeList
    
    def get_fieldsets(self, request, obj=None):
        fields = super().get_fields(request, obj)
        main_fields = ['program', 'name', 'slug', 'status', 'hidden']
        
        main = (None, {'fields': main_fields[:3] + main_fields[4:]})
        if not obj: return [main]
        
        for n in main_fields: fields.remove(n)
        main[1]['fields'] = main_fields
        
        ret = [
            main,
            ('Options', {'fields': fields}),
        ]
        responses = register_form_settings.send(obj)
        for receiver, response in responses:
            meta = get_matching_plugin(receiver.__module__)
            ret.append((meta.name, {'fields': list(response)}))
        return ret
    
    def get_readonly_fields(self, request, obj=None):
        fields = super().get_readonly_fields(request, obj)
        if obj:
            if obj.status == Form.Status.DRAFT: fields += ('status',)
            else: fields += ('program', 'slug')
        return fields
    
    def change_view(self, request, object_id, **kwargs):
        action, kwargs = None, {}
        if '_publish' in request.POST: action = 'publish'
        elif '_unpublish_confirmed' in request.POST: action = 'unpublish'
        
        if not action: return super().change_view(request, object_id, **kwargs)
        
        obj = self.get_object(request, object_id)
        getattr(obj, action)(**kwargs)
        
        return HttpResponseRedirect(request.get_full_path())
    
    def save_form(self, request, form, change):
        obj = super().save_form(request, form, change)
        if not change: return obj
        
        if obj.status == Form.Status.DRAFT:
            obj.modified = timezone.now()
        elif obj.status in (Form.Status.DISABLED, Form.Status.COMPLETED):
            if 'status' in form.changed_data: obj.completed = timezone.now()
        else: obj.completed = None
        
        return obj
        
    def response_change(self, request, obj):
        subs = []
        if obj.status != Form.Status.DRAFT:
            qs = obj.model.objects.all()
            if obj.item_model:
                qs = qs.annotate(num_items=Count('_item'))
                subs = qs.values_list('_email', 'num_items')
            else: subs = qs.values_list('_email')
        
        context = {
            **self.admin_site.each_context(request),
            'opts': self.model._meta,
            'media': self.media,
            'object': obj,
            'submissions': subs,
            'title': 'Confirmation'
        }
#        if '_publish' in request.POST:
#            return TemplateResponse(request, 'admin/publish_confirmation.html',
#                                    context)
        if '_unpublish' in request.POST:
            request.current_app = self.admin_site.name
            template_name = 'admin/formative/unpublish_confirmation.html'
            return TemplateResponse(request, template_name, context)
        
        return super().response_change(request, obj)
    
    def response_post_save_change(self, request, obj):
        app_label = self.model._meta.app_label
        url = reverse('admin:%s_formblock_formlist' % (app_label,),
                      args=(obj.id,), current_app=self.admin_site.name)
        return HttpResponseRedirect(url)


@admin.register(FormLabel, site=site)
class FormLabelAdmin(admin.ModelAdmin):
    list_display = ('path', 'style', 'form')
    list_filter = ('form',)
    
    def formfield_for_dbfield(self, db_field, **kwargs):
        formfield= super().formfield_for_dbfield(db_field, **kwargs)
        if db_field.name == 'text':
            attrs = formfield.widget.attrs
            attrs['class'], attrs['cols'] = 'vTextArea', 60
            formfield.widget = forms.Textarea(attrs=attrs)
        return formfield


class FormDependencyInline(admin.TabularInline):
    model = FormDependency
    extra = 0
    verbose_name_plural = 'dependency values'
    
    def has_add_permission(self, request, obj):
        if not obj: return False
        
        return obj.form.status == Form.Status.DRAFT
    
    def has_change_permission(self, request, obj):
        return self.has_add_permission(request, obj)
    
    def has_delete_permission(self, request, obj):
        return self.has_add_permission(request, obj)


class FormBlockBase:
    # only methods that can be safely overridden in either parent or child admin
    
    def get_fieldsets(self, request, obj=None):
        fields = list(dict.fromkeys(self.get_fields(request, obj)))
        fields.remove('dependence')
        fields.remove('negate_dependencies')
        fields.remove('no_review')
        
        main = (None, {'fields': fields})
        if not obj: return [main]
        return [
            main,
            ('Options', {'fields': ['no_review']}),
            ('Dependence', {'fields': ['dependence', 'negate_dependencies']})
        ]
    
    def get_readonly_fields(self, request, obj=None):
        fields = super().get_readonly_fields(request, obj)
        
        if obj and obj.form.status != Form.Status.DRAFT:
            fields += ('name', 'page', 'dependence', 'negate_dependencies')
        elif obj: fields += ('page',)
        
        return fields
    
    def get_form(self, request, obj=None, **kwargs):
        form = super().get_form(request, obj, **kwargs)
        
        form_id = request.GET.get('form_id')
        if form_id: form.form_id = form_id
        
        if obj and obj.form.status == Form.Status.DRAFT:
            # TODO: where would we modify form.fields instead?
            if 'dependence' in form.base_fields:
                qs = form.base_fields['dependence'].queryset
                qs = qs.filter(form=obj.form, page__gt=0)
                qs = qs.exclude(pk=obj.pk).exclude(page__gte=obj.page)
                form.base_fields['dependence'].queryset = qs
        
        return form
    
    def get_changeform_initial_data(self, request):
        changelist_filters = request.GET.get('_changelist_filters')
        if changelist_filters:
            filters = dict(parse_qsl(unquote(changelist_filters)))
            if 'page' in filters:
                return {'page': filters['page']}
        
        return {}
    
    def get_formsets_with_inlines(self, request, obj=None):
        for inline in self.get_inline_instances(request, obj):
            # only show the inline on the change form, not add:
            if not isinstance(inline, FormDependencyInline) or obj is not None:
                yield inline.get_formset(request, obj), inline
    
    def response_add(self, request, obj, **kwargs):
        if '_popup' in request.POST or '_continue' not in request.POST:
            return super().response_add(request, obj, **kwargs)
        
        opts = obj._meta
        model = opts.model_name
        if model in ('customblock', 'collectionblock'): model = 'formblock'
        url = reverse('admin:%s_%s_change' % (opts.app_label, model),
                      args=(obj.pk,), current_app=self.admin_site.name)
        
        preserved, args = request.GET.get('_changelist_filters'), {}
        if preserved: args['_changelist_filters'] = preserved
        args['form_id'] = obj.form.id
        url += '?' + urlencode(args)
        
        if 'get_url' in kwargs: return url
        return super().response_add(request, obj, post_url_continue=url)
    
    def response_change(self, request, obj):
        if '_popup' in request.POST or '_continue' not in request.POST:
            return super().response_change(request, obj)
        
        url = self.response_add(request, obj, get_url=True)
        return HttpResponseRedirect(url)
    
    def response_post_save_change(self, request, obj):
        app_label = self.model._meta.app_label
        form_id = request.GET.get('form_id')
        
        if not obj and not form_id:
            return super().response_post_save_change(request, obj)
        
        if not form_id: form_id = obj.form.id
        url = reverse('admin:%s_formblock_formlist' % (app_label,),
                      args=(form_id,), current_app=self.admin_site.name)
        changelist_filters = request.GET.get('_changelist_filters')
        if changelist_filters:
            filters = dict(parse_qsl(unquote(changelist_filters)))
            url += '?' + urlencode(filters)
        return HttpResponseRedirect(url)
    
    def response_post_save_add(self, request, obj):
        if request.GET.get('form_id'):
            return self.response_post_save_change(request, obj)
        return super().response_post_save_add(request, obj)
    
    def changeform_view(self, request, object_id=None, form_url='',
                        extra_context=None):
        form_id = request.GET.get('form_id')
        if form_id:
            form_arg = urlencode({'form_id': form_id})
            if form_url: form_url += '&' + form_arg
            else: form_url = '?' + form_arg
            
            if not extra_context: extra_context = {}
            name = get_object_or_404(Form, id=int(form_id)).name
            extra_context.update({'form_id': int(form_id), 'form_name': name})
            
            initial = self.get_changeform_initial_data(request)
            if 'page' in initial: extra_context['page'] = initial['page']
            else: extra_context['page'] = None
        
        return super().changeform_view(request, object_id, form_url,
                                       extra_context)
    
    def try_form_id(self, request, match):
        form = None
        if 'form_id' in match.kwargs: pk = match.kwargs['form_id']
        else: pk = request.GET.get('form_id')
        if not pk: return form
        
        try: form = Form.objects.get(pk=pk)
        except Form.DoesNotExist: pass
        return form
    
    def has_add_permission(self, request):
        match, app_label = request.resolver_match, self.model._meta.app_label
        # this will hide add button when we don't have the form_id
        if match and match.url_name == f'{app_label}_formblock_changelist':
            return False
        # or when the form is not a draft
        pages = (f'{app_label}_formblock_{n}' for n in ('formlist', 'change'))
        if match and match.url_name in pages:
            form = self.try_form_id(request, match)
            if form and form.status != Form.Status.DRAFT: return False
        
        return super().has_add_permission(request)
    
    def has_delete_permission(self, request, obj=None):
        return self.has_add_permission(request)


class PageListFilter(admin.SimpleListFilter):
    title = 'page'
    parameter_name = 'page'
    template = 'admin/formative/formblock/filter.html'
    
    def lookups(self, request, model_admin):
        qs = model_admin.get_queryset(request).distinct().order_by('page')
        return [ (p, f'Page {p}' if p else 'Auto-created blocks')
                 for p in qs.values_list('page', flat=True) ]
    
    def queryset(self, request, queryset):
        return queryset.filter(page=self.value())


@admin.action(description='Move to different page')
def move_blocks_action(modeladmin, request, queryset):
    if '_move' in request.POST:
        page, n = request.POST['page'], 0
        blocks, cur_page = queryset[0].form.blocks, queryset[0].page
        last_rank = blocks.filter(page=cur_page).aggregate(r=Max('_rank'))['r']
        for block in queryset.order_by('_rank'):
            block = FormBlock.objects.get(pk=block.pk) # clear side effects;
            block._rank = last_rank
            block.save() # it's now at the end of its page;
            block.page, block._rank = page, None # and then move to the new one,
            block.save() # so that the others on the old page are left sorted
            n, last_rank = n + 1, last_rank - 1
        
        msg = f'Moved {n} blocks to page {page}'
        modeladmin.message_user(request, msg, messages.SUCCESS)
        
        return HttpResponseRedirect(request.get_full_path())
    
    template_name = 'admin/formative/move_page.html'
    request.current_app = site.name
    min_page = max(block.min_allowed_page() for block in queryset)
    last_page = queryset[0].form.blocks.aggregate(p=Max('page'))['p']
    max_page = min(block.max_allowed_page(last_page) for block in queryset)
    new = max_page == last_page
    context = {
        **site.each_context(request),
        'opts': modeladmin.model._meta,
        'media': modeladmin.media,
        'blocks': queryset,
        'movable': queryset[0].form.status == Form.Status.DRAFT,
        'form': MoveBlocksAdminForm(max_page, min_page=min_page, new_page=new),
        'title': 'Move Blocks'
    }
    return TemplateResponse(request, template_name, context)

@admin.register(FormBlock, site=site)
class FormBlockAdmin(FormBlockBase, PolymorphicParentModelAdmin,
                     DynamicArrayMixin):
    child_models = (FormBlock, CustomBlock, CollectionBlock)
    list_display = ('_rank', 'name', 'block_type', 'dependence', 'labels_link')
    list_display_links = ('name',)
    list_editable = ('_rank',)
    list_filter = (PageListFilter,)
    form = StockBlockAdminForm
    inlines = [FormDependencyInline]
    actions = [move_blocks_action]
    sortable_by = ()
    polymorphic_list = True
    
    @admin.display(description='labels')
    def labels_link(self, obj):
        app_label = self.model._meta.app_label
        url = reverse('admin:%s_formlabel_changelist' % (app_label,),
                      current_app=self.admin_site.name)
        url += f'?form__id__exact={obj.form.pk}&path__startswith={obj.name}'
        return format_html('<a href="{}">{} labels</a>', url, obj.num_labels)
    
    @admin.display(description='type')
    def type(self, obj):
        # read-only version of the StockBlockAdminForm type field
        return obj.options['type']
    
    def get_fieldsets(self, request, obj=None):
        fieldsets = super().get_fieldsets(request, obj)
        fields = fieldsets[0][1]['fields']
        
        if not obj: return fieldsets
        
        options = fieldsets[1][1]['fields']
        admin_fields = obj.stock.admin_fields().keys()
        readonly_fields = obj.stock.admin_published_readonly()
        if obj.form.status == Form.Status.DRAFT:
            # special case: not really an option, so it goes on the general tab
            if 'choices' in admin_fields: fields.append('choices')
            options += [ f for f in admin_fields if f != 'choices' ]
        else:
            options += [ f for f in admin_fields if f not in readonly_fields ]
        
        sets = [(None, {'fields': fields}), ('Options', {'fields': options})] 
        return sets + fieldsets[2:]
    
    def get_readonly_fields(self, request, obj=None):
        fields = super().get_readonly_fields(request, obj)
        
        if not obj: return fields
        fields += ('type',)
        
        if obj.form.status != Form.Status.DRAFT:
            for field, label in obj.stock.admin_published_readonly().items():
                def field_callable(f, l):
                    @admin.display(description=l)
                    def callable(self, obj):
                        if f not in obj.options: return '-'
                        return obj.options[f]
                    return callable
                
                setattr(self, field,
                        types.MethodType(field_callable(field, label), self))
                fields += (field,)
        
        return fields
    
    def get_urls(self):
        urls = super().get_urls()
        
        url = path('form/<int:form_id>/', self.formlist_view,
                   name='%s_formblock_formlist' % (self.model._meta.app_label,))
        return [url] + urls
    
    def formlist_view(self, request, form_id, **kwargs):
        page = request.GET.get('page')
        if not page: return HttpResponseRedirect(request.path + '?page=1')
        
        name = get_object_or_404(Form, id=int(form_id)).name
        context = {'form_id': form_id, 'form_name': name}
        return self.changelist_view(request, extra_context=context)
    
    def get_queryset(self, request):
        qs = super().get_queryset(request)
        match, app_label = request.resolver_match, self.model._meta.app_label
        if match and match.url_name == f'{app_label}_formblock_formlist':
            qs = qs.filter(form_id=match.kwargs['form_id'])
        
        path_filter = Q(form__label__path__startswith=F('name'))
        qs = qs.annotate(num_labels=Count('form__label', filter=path_filter))
        return qs
    
    def get_changelist_form(self, request, **kwargs):
        class HiddenWithHandleInput(forms.HiddenInput):
            template_name = 'admin/formative/widgets/hidden_with_handle.html'
        
        kwargs['widgets'] = {'_rank': HiddenWithHandleInput}
        return super().get_changelist_form(request, **kwargs)
    
    def change_view(self, *args, **kwargs):
        # we still have the bulk action for delete - it redirects properly
        kwargs['extra_context'] = {'show_delete': False,
                                   'show_save_and_add_another': False}
        return super().change_view(*args, **kwargs)
    
    def get_preserved_filters(self, request):
        match, app_label = request.resolver_match, self.model._meta.app_label
        
        # we have to reimplement, because of our custom changelist view
        if self.preserve_filters and match:
            current_url = f'{match.app_name}:{match.url_name}'
            changelists = [ f'admin:{app_label}_formblock_{n}'
                            for n in ('formlist', 'changelist') ]
            if current_url in changelists: preserved = request.GET.urlencode()
            else: preserved = request.GET.get('_changelist_filters')
            
            args = {}
            if preserved: args['_changelist_filters'] = preserved
            if current_url == changelists[0]:
                args['form_id'] = match.kwargs['form_id']
            return urlencode(args)
        return ''


class FormBlockChildAdmin(FormBlockBase, PolymorphicChildModelAdmin):
    base_model = FormBlock
    inlines = [FormDependencyInline]


@admin.register(CustomBlock, site=site)
class CustomBlockAdmin(FormBlockChildAdmin, DynamicArrayMixin):
    form = CustomBlockAdminForm
    radio_fields = {'type': admin.VERTICAL}
    
    def choices(self, obj):
        return obj.options['choices']
    
    def get_fieldsets(self, request, obj=None):
        fieldsets = super().get_fieldsets(request, obj)
        fields = fieldsets[0][1]['fields']
        names = ['type', 'required', 'num_lines', 'min_chars', 'max_chars',
                 'min_words', 'max_words']
        
        if not obj:
            return [(None, {'fields': ['name', 'page', 'type']})]
        
        options = fieldsets[1][1]['fields']
        if obj.type == CustomBlock.InputType.TEXT: add = names[:1] + names[2:]
        elif obj.dependence: options.append('default_value')
        if obj.type == CustomBlock.InputType.BOOLEAN: add = names[:1]
        elif obj.type == CustomBlock.InputType.CHOICE:
            add = names[:2] + ['choices']
        elif obj.type == CustomBlock.InputType.NUMERIC:
            add = names[:2]
            options += ['numeric_min', 'numeric_max']
        
        names += ['choices', 'numeric_min', 'numeric_max', 'default_value']
        main = [ f for f in fields if f not in names ] + add
        
        sets = [(None, {'fields': main}), ('Options', {'fields': options})]
        return sets + fieldsets[2:]
    
    def get_readonly_fields(self, request, obj=None):
        fields = super().get_readonly_fields(request, obj)
        
        if obj: fields += ('type',)
        if obj and obj.form.status != Form.Status.DRAFT:
            if obj.type == CustomBlock.InputType.TEXT:
                fields += ('num_lines', 'max_chars')
            elif obj.type == CustomBlock.InputType.CHOICE:
                fields += ('choices',)
        return fields


@admin.register(CollectionBlock, site=site)
class CollectionBlockAdmin(FormBlockChildAdmin, DynamicArrayMixin):
    form = CollectionBlockAdminForm
    
    def get_fieldsets(self, request, obj=None):
        fieldsets = super().get_fieldsets(request, obj)
        fields = fieldsets[0][1]['fields']
        names = ['fixed', 'name1', 'name2', 'name3', 'has_file',
                 'min_items', 'max_items', 'file_optional']
        
        if not obj: add = names[:5]
        elif obj.fixed: add = names[:4] + ['choices']
        elif not obj.has_file: add = names[:6]
        else: add = names[:]
        
        main = ['name', 'page'] + add
        if not obj: return [(None, {'fields': main})]
        
        options = fieldsets[1][1]['fields']
        names += ['name', 'page', 'file_types', 'max_filesize',
                  'autoinit_filename', 'choices', 'wide']
        options += [ f for f in fields if f not in names ]
        if obj.has_file:
            options += ['file_types', 'max_filesize', 'autoinit_filename']
            total, processing = False, {}
            for name in obj.allowed_filetypes() or []:
                filetype = FileType.by_type(name)()
                if filetype.admin_limit_fields(): options.append(name)
                if filetype.admin_processing_fields(): processing[name] = True
                if filetype.admin_total_fields(): total = True
            if total: options.append('total')
            for name in processing: options.append(name + '_proc')
        
        if obj.name1 and obj.name2: options.append('wide')
        
        sets = [(None, {'fields': main}), ('Options', {'fields': options})]
        return sets + fieldsets[2:]
    
    def get_readonly_fields(self, request, obj=None):
        fields = super().get_readonly_fields(request, obj)
        
        if obj: fields += ('fixed', 'has_file')
        if obj and obj.form.status != Form.Status.DRAFT:
            fields += ('name1', 'name2', 'name3')
        
        return fields


class SubmittedListFilter(admin.SimpleListFilter):
    title = 'status'
    parameter_name = '_submitted'
    
    def lookups(self, request, model_admin):
        return (('yes', 'submitted'), ('no', 'unsubmitted'))
    
    def queryset(self, request, queryset):
        if self.value() == 'yes': return queryset.exclude(_submitted=None)
        if self.value() == 'no': return queryset.filter(_submitted=None)


EMAILS_PER_SECOND = 10

@admin.action(description='Send an email to applicants')
def send_email_action(modeladmin, request, queryset):
    if '_send' in request.POST:
        # TODO: celery task
        subject = Template(request.POST['subject'])
        content = Template(request.POST['content'])
        
        iterator = queryset.iterator()
        batch, form, last_time, done, n = [], None, None, False, 0
        while not done:
            submission = next(iterator, None)
            if not submission: done = True
            else: batch.append(submission)
            
            if len(batch) == EMAILS_PER_SECOND or done:
                if last_time:
                    this_time = time.time()
                    remaining = last_time + 1 - this_time
                    if remaining > 0: time.sleep(remaining)
                last_time = time.time()
                
                with mail.get_connection() as conn:
                    for sub in batch:
                        if not form: form = sub._get_form()
                        context = {
                            'submission': sub, 'form': form,
                            'submission_link': submission_link(sub, form)
                        }
                        if sub._submitted: sub._update_context(form, context)
                        
                        n += 1
                        send_email(content, sub._email, subject,
                                   context=context, connection=conn)
                batch = []
        
        msg = f'Emails sent to {n} recipients.'
        modeladmin.message_user(request, msg, messages.SUCCESS)
        
        return HttpResponseRedirect(request.get_full_path())
    
    form = queryset[0]._get_form()
    template_name = 'admin/formative/email_applicants.html'
    request.current_app = site.name
    context = {
        **site.each_context(request),
        'opts': modeladmin.model._meta,
        'media': modeladmin.media,
        'submissions': queryset,
        'form': EmailAdminForm(form=form),
        'email_templates': form.email_templates(),
        'title': 'Email Applicants'
    }
    return TemplateResponse(request, template_name, context)
    

class SubmissionAdmin(admin.ModelAdmin):
    list_display = ('_email', '_created', '_modified', '_submitted')
    list_filter = ('_email', SubmittedListFilter)
    readonly_fields = ('_submitted', 'items_index',)
    form = SubmissionAdminForm
    actions = [send_email_action]
    
    @admin.display(description='items')
    def items_index(self, obj):
        app, name = self.model._meta.app_label, self.model._meta.model_name
        args = f'?_submission___id__exact={obj._id}'
        try:
            url = reverse('admin:%s_%s_changelist' % (app, name + '_i'),
                          current_app=self.admin_site.name) + args
        except NoReverseMatch: return ''
        
        return mark_safe(f'<a href="{url}">items listing</a>')
    
    def change_view(self, request, object_id, **kwargs):
        action, kwargs = None, {}
        if '_submit_confirmed' in request.POST: action = 'submit'
        elif '_unsubmit_confirmed' in request.POST: action = 'unsubmit'
        
        if not action:
            kwargs['extra_context'] = {'model_is_submission': True}
            return super().change_view(request, object_id, **kwargs)
        
        obj = self.get_object(request, object_id)
        rec, rtype = None, SubmissionRecord.RecordType.SUBMISSION
        try: rec = SubmissionRecord.objects.get(submission=obj._id, type=rtype)
        except SubmissionRecord.DoesNotExist: pass
        
        if action == 'submit':
            if not rec: obj._get_form().submit_submission(obj)
            else:
                obj._submit()
                rec.deleted = False
        else:
            obj._submitted = None
            obj.save()
            if rec: rec.deleted = True
        if rec: rec.save()
        return HttpResponseRedirect(request.get_full_path())
    
    def response_change(self, request, obj):
        context = {
            **self.admin_site.each_context(request),
            'opts': self.model._meta,
            'media': self.media,
            'object': obj,
            'title': 'Confirmation'
        }
        template_name = 'admin/formative/submit_confirmation.html'
        if '_submit' in request.POST or '_unsubmit' in request.POST:
            request.current_app = self.admin_site.name
            if '_submit' in request.POST: context['submit'] = True
            else: context['unsubmit'] = True
            
            return TemplateResponse(request, template_name, context)
        
        return super().response_change(request, obj)
    
    def view_on_site(self, obj):
        url = obj._get_absolute_url()
        if obj._submitted: url += 'review'
        return url


class SubmissionItemAdmin(admin.ModelAdmin):
    list_display = ('_id', '_submission', '_collection', '_rank', '_file')
    list_filter = (
        '_submission', '_collection',
        ('_file', admin.EmptyFieldListFilter)
    )
    readonly_fields = ('_submission', '_block', '_rank')
    ordering = ('_submission', '_collection', '_block', '_rank')
    form = SubmissionItemAdminForm
    
    def has_add_permission(self, request):
        return False



from django.apps import apps
from django.db.models import Model, Q
from django.conf import settings
from django.core import mail
from django.template import Context, Template, loader
from django.utils.translation import gettext_lazy as _
from django.contrib import admin
import os, glob
from pathlib import Path


def create_model(name, fields, app_label='formative', module='',
                 program=None, meta=None, base_class=Model):
    class Meta:
        pass

    setattr(Meta, 'app_label', app_label)
    if meta is not None:
        for key, value in meta.__dict__.items():
            if key[:2] == '__' or key == 'abstract': continue
            setattr(Meta, key, value)
    setattr(Meta, 'db_table', name)
    
    if not module: module = app_label
    attrs = {'__module__': module, 'Meta': Meta}
    attrs.update(dict(fields))

    # Create the class, which automatically triggers ModelBase processing
    model = type(name, (base_class,), attrs)
    
    if program: model._meta.program_slug = program
    return model

def remove_p(text):
    s = text.strip()
    if s[-3-1:] == '</p>':
        i = s.rindex('<p>')
        return s[i+3:-3-1]
    
    return text

def send_email(template, to, subject, context={}, connection=None):
    new_context = { 'settings': settings }
    new_context.update(context)
    context = Context(new_context)
    
    if type(template) != Template: context = new_context # wtf, Django
    
    sub = ' '.join(subject.render(context).splitlines()).rstrip()
    message = template.render(context)
    
    email = mail.EmailMessage(sub, message, settings.CONTACT_EMAIL, [to],
                              connection=connection)
    return email.send()

def submission_link(s, form, rest=''):
    server = settings.DJANGO_SERVER
    if ':' in server or server.endswith('.local'): proto = 'http'
    else: proto = 'https'
    
    return f'{proto}://{server}/{form.program.slug}/{form.slug}/{s._id}/{rest}'

def get_file_extension(name):
    return Path(name).suffix[1:].lower()

def thumbnail_path(path, ext=None):
    idx = path.rindex('.')
    return path[:idx] + '_tn' + (ext and '.'+ext or path[idx:])

def subtitle_path(path, lang):
    idx = path.rindex('.')
    return path[:idx] + '_s_' + lang + '.vtt'

def delete_file(file):
    if os.path.isfile(file.path): os.remove(file.path)
    
    thumb = thumbnail_path(file.path)
    if os.path.isfile(thumb): os.remove(thumb)
    
    for path in glob.glob(subtitle_path(file.path, '*')):
        os.remove(path)

def human_readable_filesize(size, decimal_places=2):
    for unit in ['bytes', 'kB', 'MB', 'GB', 'TB', 'PB']:
        if size < 1024 or unit == 'PB': break
        size /= 1024
    return f"{size:.{decimal_places}f} {unit}"

def any_name_field(**kwargs):
    Qs = [ Q(**{ namen + (k != '_' and k or ''): v for k, v in kwargs.items() })
           for namen in ('name1', 'name2', 'name3') ]
    return Qs[0] | Qs[1] | Qs[2]

def get_tooltips():
    return {
        'previoustip': _('Previous Page'),
#        'sortabletip': _('Drag to reorder'),
#        'uploadtip': _('Replace File'),
    }



from django.apps import AppConfig
from django.db.models import F, Value, CharField
from django.db.models.functions import Concat


class FormativeConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'formative'
    
    def get_model(self, model_name, require_ready=True):
        model = None
        try: model = super().get_model(model_name, require_ready)
        except LookupError: pass
        if model or not require_ready or '_' not in model_name: return model
        
        from .models import Form
        part = model_name[:model_name.index('_')]
        queryset = Form.objects.filter(program__db_slug__startswith=part)
        name = Concat(F('program__db_slug'), Value('_'), F('db_slug'),
                      output_field=CharField())
        for form in queryset.annotate(n=name).filter(n=model_name):
            # accessing the models will register them for the call to super
            form.model, form.item_model
        
        return super().get_model(model_name)
            
    def ready(self):
        from . import signals



from django.db import models

from . import StockWidget

class EmailWidget(StockWidget):
    TYPE = 'email'
    
    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)
        
        self.template_name = 'generic.html'
        self.review_template_name = 'review.html'
    
    def fields(self):
        field = models.EmailField(blank=True, null=True)
        if self.name == 'email':
            # only the validation block can use the name 'email' - needs unique
            field = models.EmailField(blank=True, unique=True)
        
        return [(self.field_name(), field)]



from django import template, forms
from collections import OrderedDict

from ..models import FormLabel


register = template.Library()

# get the form field corresponding to the given block
@register.simple_tag
def block_field(form, block):
    if block.name in form.fields: return form[block.name]
    return None

# get the inline formset corresponding to the given collection block
@register.simple_tag
def block_formset(formsets, block):
    if block.pk in formsets: return formsets[block.pk]
    return None

@register.simple_tag
def block_labels(labels, block):
    if block.block_type() == 'collection':
        key = f'{block.name}{block.id}_'
        if key in labels: return labels[key]
    
    if block.name in labels: return labels[block.name]
    return {}

@register.simple_tag
def collection_items(items, block):
    if block.pk in items: return items[block.pk]
    return None

@register.simple_tag
def item_columns(item, block, uploading):
    if not item: return { 'fields': True }
    
    return {
        'fields': not item._error and not uploading,
        'progress': not item._error and uploading,
        'message': item._error,
        'upload': item._error or block.file_optional
    }

@register.simple_tag
def item_form(formset, item):
    if type(item) == int: return formset[item-1]
    
    for form in formset:
        if form.instance == item: return form
    return None

@register.simple_tag
def form_hidden(form, name):
    if form: return form['_' + name]
    return None

@register.filter
def get_by_field(errors, name):
    for form_errors in errors:
        if name in form_errors: return form_errors[name][0]
    return ''

@register.filter
def get_by_style(labels, style):
    if labels and style in labels: return labels[style]
    return None

@register.filter
def get_by_choice(labels, choice):
    if choice in labels: return labels[choice]
    return None

@register.filter
def closest_label(labels):
    names = ('WIDGET', 'HORIZONTAL', 'VERTICAL')
    styles = [ FormLabel.LabelStyle[n] for n in names ]
    for style in styles:
        label = get_by_style(labels, style)
        if label: return label
    return None

@register.filter
def for_choice_value(labels, value):
    if value in labels: return closest_label(labels[value])
    return None

@register.filter
def for_item_field(labels, field):
    if field in labels: return labels[field]
    return None

@register.simple_tag
def unbound_checkbox_field(form_block, name, radio=False):
    anon_form = forms.Form()
    if radio: field = forms.ChoiceField(choices=[(name, name)],
                                        widget=forms.RadioSelect)
    else: field = forms.BooleanField(widget=forms.CheckboxInput)
    anon_form.fields['_' + form_block.stock.field_name(name)] = field
    field.required = False
    return anon_form

@register.simple_tag
def unbound_radio_field(form_block, name):
    return unbound_checkbox_field(form_block, name, radio=True)

@register.simple_tag(takes_context=True)
def include_stock(context, block, labels, review=False):
    stock = block.stock
    
    name = review and stock.review_template_name or stock.template_name
    template = context.template.engine.get_template('formative/stock/' + name)
    
    fields = []
    for n in stock.widget_names():
         field_name = stock.field_name(n)
         if field_name not in context['form'].fields: return ''
         fields.append((n, context['form'][field_name]))
    fields_dict = OrderedDict(fields)
    
    return template.render(context.new({
        'form_block': block,
        'block_fields': fields_dict,
        'labels': labels
    }))

@register.simple_tag(takes_context=True)
def include_stock_review(context, block, labels):
    return include_stock(context, block, labels, review=True)



from django import forms
from django.core.exceptions import ValidationError
from django.contrib.admin import widgets
from copy import deepcopy
from django_better_admin_arrayfield.forms.fields import DynamicArrayField
from django_better_admin_arrayfield.forms.widgets import DynamicArrayWidget

import datetime

from ..signals import register_program_settings, register_form_settings
from ..filetype import FileType
from ..stock import StockWidget
from ..models import Program, Form, FormBlock, CustomBlock, CollectionBlock, \
    Submission, SubmissionItem
from ..validators import validate_program_identifier, validate_form_identifier,\
    validate_formblock_identifier


class NullWidget(forms.Widget):
    @property
    def is_hidden(self): return True
    
    def value_omitted_from_data(self, data, files, name): return False
    
    def render(self, name, value, **kwargs): return ''


class JSONPseudoBoundField(forms.BoundField):
    @property
    def data(self):
        # always start with initial value; other form fields will modify
        return self.initial


class JSONPseudoField(forms.Field):
    widget = NullWidget
    
    def get_bound_field(self, form, field_name):
        return JSONPseudoBoundField(form, self, field_name)


class NegatedBooleanField(forms.BooleanField):
    def prepare_value(self, value): return not value
    
    def to_python(self, value): return not super().to_python(value)
    
    def has_changed(self, initial, data):
        # something unusual about the forms API is making this backwards
        return (not self.to_python(initial)) != self.to_python(data)


DISPLAY_MAX = 9999999999

class SplitDictWidget(forms.MultiWidget):
    template_name = 'admin/formative/widgets/split_dict.html'
    
    def __init__(self, fields, two_column=False, attrs=None):
        self.fields, self.two_column = fields, two_column
        
        subwidgets = {}
        for name, field in fields.items():
            widget = forms.TextInput(attrs=attrs)
            if isinstance(field, forms.IntegerField):
                if not attrs: int_attrs = {}
                else: int_attrs = attrs.copy()
                
                int_attrs['min'], int_attrs['max'] = 0, DISPLAY_MAX
                if name.startswith('min_'):
                    int_attrs['placeholder'] = 'minimum'
                elif name.startswith('max_'):
                    int_attrs['placeholder'] = 'maximum'
                
                widget = widgets.AdminIntegerFieldWidget(attrs=int_attrs)
            subwidgets[name] = widget
        
        super().__init__(subwidgets)
    
    def decompress(self, val):
        if val:
            ret = []
            for name, field in self.fields.items():
                if name in val: ret.append(val[name])
                else: ret.append(None)
            return ret
        return [None] * len(self.fields)
    
    def get_context(self, name, value, attrs):
        context = super().get_context(name, value, attrs)
        context['two_column'] = self.two_column
        for widget in context['widget']['subwidgets']:
            name = widget['name'][widget['name'].index('_')+1:]
            if name.startswith('min_') or name.startswith('max_'):
                name = name[4:]
            if name.startswith('proc_'): name = name[5:]
            widget['short_name'] = name
        
        return context


class SplitDictField(forms.MultiValueField):
    widget = SplitDictWidget
    
    def __init__(self, fields, **kwargs):
        self.fields_dict = fields
        super().__init__(fields.values(), **kwargs)
    
    def compress(self, data):
        if not data: return None
        ret = {}
        for i, (name, field) in enumerate(self.fields_dict.items()):
            if data[i] is not None: ret[name] = data[i]
        if ret: return ret
        return None


class SingletonChoiceField(forms.ChoiceField):
    def prepare_value(self, value):
        if not value: return None
        return value[0]
    
    def to_python(self, value):
        if not value: return None
        return [super().to_python(value)]
    
    def validate(self, value):
        if value: value = value[0]
        super().validate(value)


class TypedCharField(forms.CharField):
    def __init__(self, coerce=lambda val: val, **kwargs):
        self.coerce = coerce
        self.msg = 'There was an error.'
        super().__init__(**kwargs)
    
    def coerce_value(self, value):
        if value is None or value == '': return None
        if type(value) != str: return value
        
        try: value = self.coerce(value)
        except (ValueError, TypeError, KeyError):
            raise ValidationError(self.msg)
        
        return value
    
    def clean(self, value):
        value = super().clean(value)
        return self.coerce_value(value)
    
    def has_changed(self, initial, data):
        if not super().has_changed(initial, data): return False
        try: ret = self.coerce_value(data) != self.coerce_value(initial)
        except ValidationError: return True
        return ret


class JSONDateTimeWidget(widgets.AdminSplitDateTime):
    def decompress(self, val):
        if not val: return super().decompress(val)
        
        return super().decompress(datetime.datetime.fromisoformat(val))


class JSONDateTimeField(forms.SplitDateTimeField):
    widget = JSONDateTimeWidget
    
    def compress(self, data):
        val = super().compress(data)
        if not val: return val
        
        return val.isoformat()


class AdminJSONFormMetaclass(forms.models.ModelFormMetaclass):
    def __new__(cls, class_name, bases, attrs):
        base_formfield_callback, base_static_fields = None, None
        for b in bases:
            if hasattr(b, 'Meta') and hasattr(b.Meta, 'formfield_callback'):
                base_formfield_callback = b.Meta.formfield_callback
                break
        for b in bases:
            if hasattr(b, 'Meta') and hasattr(b.Meta, 'static_fields'):
                base_static_fields = b.Meta.static_fields
                break
        callback = attrs.pop('formfield_callback', base_formfield_callback)
        
        json_fields, fields, dynamic_fields, static_fields = {}, [], [], []
        if 'Meta' in attrs:
            json_fields = deepcopy(getattr(attrs['Meta'], 'json_fields', {}))
            fields = getattr(attrs['Meta'], 'fields', [])
            if fields != forms.ALL_FIELDS: fields = list(fields)
            dynamic_fields = getattr(attrs['Meta'], 'dynamic_fields', False)
            static_fields = list(getattr(attrs['Meta'], 'static_fields', []))
        
        def json_field_callback(db_field, **kwargs):
            if db_field.name not in json_fields.keys():
                if not callback: return db_field.formfield(**kwargs)
                return callback(db_field, **kwargs)
            return JSONPseudoField(required=False)
        
        if json_fields:
            attrs['formfield_callback'] = json_field_callback
            exclude_fields = []
            if dynamic_fields:
                # Meta.static_fields must be used for dynamic_fields forms
                init_fields = static_fields
                if not init_fields: init_fields = base_static_fields
                
                for f in init_fields and fields or []:
                    # requested fields not in base's list are dynamic ones
                    if f not in init_fields:
                        # non-model fields have to be statically declared
                        exclude_fields.append(f) # remove; added back, in init
            
            if fields and fields != forms.ALL_FIELDS:
                fields = [ f for f in fields
                           if f not in json_fields and f not in exclude_fields ]
                json_keys = {}
                for field in json_fields:
                    if '.' in field: val = field[:field.index('.')]
                    else: val = field
                    json_keys[val] = True
                attrs['Meta'].fields = list(json_keys) + fields
        
        new_class = super().__new__(cls, class_name, bases, attrs)
        
        new_class._meta.json_fields = json_fields
        new_class._meta.static_fields = static_fields
        return new_class


class AdminJSONForm(forms.ModelForm, metaclass=AdminJSONFormMetaclass):
    def __init__(self, *args, admin_fields=None, **kwargs):
        self.admin_fields = admin_fields
        
        if 'instance' in kwargs and kwargs['instance']:
            instance, initial = kwargs['instance'], {}
            if 'initial' in kwargs: initial = kwargs['initial']
            
            for name, fields in self._meta.json_fields.items():
                if '.' in name:
                    p1, p2 = name[:name.index('.')], name[name.index('.')+1:]
                    json = getattr(instance, p1)
                    if p2 not in json: value = {}
                    else: value = json[p2]
                else: value = getattr(instance, name)
                
                for field in fields:
                    json_field = field
                    # TODO: might need to generalize:
                    if field.endswith('_proc'): json_field = field[:-5]
                    if json_field in value: initial[field] = value[json_field]
            
            kwargs['initial'] = initial
        
        super().__init__(*args, **kwargs)
        
        if self.admin_fields:
            for name, field in self.admin_fields.items():
                self.fields[name] = field
    
    def clean(self):
        super().clean()
        cleaned_data = self.cleaned_data
        
        for name, fields in self._meta.json_fields.items():
            base, part = name, None
            if '.' in name: base = name[:name.index('.')]
            
            if not cleaned_data[base]: cleaned_data[base] = {}
            dest = cleaned_data[base]
            if '.' in name:
                part = name[name.index('.')+1:]
                if part not in dest: dest[part] = {}
                dest = dest[part]
            
            for field in fields:
                if field not in cleaned_data: continue
                json_field = field
                if field.endswith('_proc'): json_field = field[:-5]
                
                test = bool
                if isinstance(self.fields[field], forms.BooleanField): pass
                elif isinstance(self.fields[field], TypedCharField):
                    test = lambda x: x is not None
                elif isinstance(self.fields[field], forms.CharField): pass
                elif isinstance(self.fields[field], forms.ChoiceField): pass
                elif isinstance(self.fields[field], SplitDictField): pass
                else: test = lambda x: x is not None
                
                if test(cleaned_data[field]):
                    dest[json_field] = cleaned_data[field]
                else: dest.pop(json_field, None)
            
            if '.' in name and not dest: cleaned_data[base].pop(part, None)
        
        return cleaned_data


class ProgramAdminForm(AdminJSONForm):
    name = forms.CharField(
        max_length=Program._meta.get_field('name').max_length,
        widget=widgets.AdminTextInputWidget,
        help_text='Full name, displayed on the listing of forms.'
    )
    slug = forms.SlugField(
        label='identifier', allow_unicode=True, max_length=30,
        widget=forms.TextInput(attrs={'size': 32}),
        validators=[validate_program_identifier],
        help_text='A short name to uniquely identify the program. '
                  'Cannot be changed.'
    )
    home_url = forms.URLField(
        required=False,
        help_text='URL for the home icon link, when viewing this program.'
    )
    
    class Meta:
        static_fields = ('name', 'slug', 'description', 'hidden', 'home_url')
        json_fields = {'options': ['home_url']}
        dynamic_fields = True
        widgets = {
            'description': widgets.AdminTextareaWidget(attrs={'rows': 3})
        }
    
    def __init__(self, *args, **kwargs):
        if 'instance' in kwargs and kwargs['instance']:
            responses = register_program_settings.send(self)
            admin_fields = { k: v for _, r in responses for k, v in r.items() }
            self._meta.json_fields['options'] += list(admin_fields.keys())
            kwargs['admin_fields'] = admin_fields
        
        super().__init__(*args, **kwargs)


class FormAdminForm(AdminJSONForm):
    name = forms.CharField(
        max_length=Form._meta.get_field('name').max_length,
        widget=widgets.AdminTextInputWidget,
        help_text='Full name, displayed on the form.'
    )
    slug = forms.SlugField(
        label='identifier', allow_unicode=True, max_length=30,
        widget=forms.TextInput(attrs={'size': 32}),
        validators=[validate_form_identifier],
        help_text='A short name to uniquely identify the form. '
                  'Cannot be changed after publishing.'
    )
    status = forms.ChoiceField(
        choices=Form.Status.choices,
        widget=widgets.AdminRadioSelect(attrs={'class': 'radiolist'})
    )
    hidden = forms.BooleanField(required=False)
    access_enable = forms.CharField(required=False, label='?access= password')
    review_pre = forms.CharField(
        required=False, widget=widgets.AdminTextareaWidget(attrs={'rows': 5}),
        label='pre review text'
    )
    review_post = forms.CharField(
        required=False, widget=widgets.AdminTextareaWidget(attrs={'rows': 5}),
        label='post review text'
    )
    submitted_review_pre = forms.CharField(
        required=False, widget=widgets.AdminTextareaWidget(attrs={'rows': 5}),
        label='submitted pre review text'
    )
    timed_completion = JSONDateTimeField(required=False)
    complete_submit_time = forms.IntegerField(
        required=False, widget=widgets.AdminIntegerFieldWidget,
        help_text='Extra time to complete uploads/submission, in minutes. ' \
                  'Default is 5.'
    )
    no_review_after_submit = forms.BooleanField(required=False)
    thanks = forms.CharField(
        required=False, widget=widgets.AdminTextareaWidget(attrs={'rows': 5}),
        label='thanks page text'
    )
    
    class Meta:
        static_fields = ('program', 'name', 'slug', 'status', 'hidden')
        json_fields = {'options': [
            'hidden', 'access_enable', 'review_pre', 'review_post',
            'submitted_review_pre', 'timed_completion', 'complete_submit_time',
            'no_review_after_submit', 'thanks'
        ]}
        dynamic_fields = True
    
    def __init__(self, *args, **kwargs):
        program_form = None
        if 'instance' in kwargs and kwargs['instance']:
            program_form = kwargs['instance']
            
            responses = register_form_settings.send(program_form)
            admin_fields = { k: v for _, r in responses for k, v in r.items() }
            self._meta.json_fields['options'] += list(admin_fields.keys())
            kwargs['admin_fields'] = admin_fields
            
        super().__init__(*args, **kwargs)
        
        if program_form and program_form.status != Form.Status.DRAFT:
            self.fields['status'].choices = self.fields['status'].choices[1:]
        
        if not program_form:
            del self.fields['status']
            for n in ('access_enable', 'review_pre', 'review_post', 'thanks',
                      'submitted_review_pre', 'timed_completion',
                      'complete_submit_time', 'no_review_after_submit'):
                del self.fields[n]


class FormBlockAdminForm(forms.ModelForm):
    name = forms.SlugField(
        label='identifier', allow_unicode=True, max_length=32,
        widget=forms.TextInput(attrs={'size': 34}),
        validators=[validate_formblock_identifier],
        help_text='A short name to identify the block, not displayed publicly.'
    )
    
    class Meta:
        model = FormBlock
        fields = ('name', 'page', 'dependence', 'negate_dependencies')
    
    def validate_unique(self):
        # FormBlockAdmin.get_form sets this; it's needed for validation and save
        self.instance.form_id = self.form_id
        
        super().validate_unique()


def stock_type_display_name(name): # we need a default display name
    # plugins can prefix a type name with "pluginname." to avoid collisions
    return name[name.rfind('.')+1:]


class StockBlockAdminForm(FormBlockAdminForm, AdminJSONForm):
    type = forms.ChoiceField(choices=[ (n, stock_type_display_name(n))
                                       for n in StockWidget.types.keys() ],
                             widget=widgets.AdminRadioSelect)
    no_review = NegatedBooleanField(label='show in review', required=False)
    
    class Meta:
        exclude = ('form', '_rank')
        static_fields = ('name', 'page', 'type', 'no_review',
                         'dependence', 'negate_dependencies')
        json_fields = {'options': ['type', 'no_review']}
        dynamic_fields = True
    
    def __init__(self, *args, **kwargs):
        stock = None
        if 'instance' in kwargs and kwargs['instance']:
            stock = kwargs['instance'].stock
            admin_fields = stock.admin_fields()
            if kwargs['instance'].form.status != Form.Status.DRAFT:
                for n in stock.admin_published_readonly(): admin_fields.pop(n)
            # safe to modify the values of json_fields, just not the keys:
            self._meta.json_fields['options'] += list(admin_fields.keys())
            kwargs['admin_fields'] = admin_fields
        
        super().__init__(*args, **kwargs)
        
        if not stock: del self.fields['no_review']
    
    def clean(self):
        super().clean()
        cleaned_data = self.cleaned_data
        if not self.admin_fields: return
        
        data = {}
        for field in self.admin_fields:
            if field in cleaned_data: data[field] = cleaned_data[field]
        fields = self.instance.stock.admin_clean(data)
        
        err = False
        for field, val in fields.items():
            if isinstance(val, ValidationError):
                self.add_error(field, val)
                err = True
        if not err:
            for field, val in fields.items():
                self.cleaned_data[field] = val


class CustomBlockAdminForm(FormBlockAdminForm, AdminJSONForm):
    no_review = NegatedBooleanField(label='show in review', required=False)
    choices = DynamicArrayField(
        forms.CharField(max_length=CustomBlock.CHOICE_VAL_MAXLEN),
    )
    numeric_min = forms.IntegerField(required=False)
    numeric_max = forms.IntegerField(required=False)
    default_value = TypedCharField(
        required=False, widget=widgets.AdminTextInputWidget,
        help_text="Value that will be used if the block isn't shown " \
                  "because the dependence is not met."
    )
    
    class Meta:
        exclude = ('form', '_rank')
        json_fields = {'options': ['no_review', 'choices', 'default_value',
                                   'numeric_min', 'numeric_max']}
    
    def __init__(self, *args, **kwargs):
        block = None
        if 'instance' in kwargs and kwargs['instance']:
            block = kwargs['instance']
        
        super().__init__(*args, **kwargs)
        
        if not block:
            del self.fields['numeric_min'], self.fields['numeric_max']
            del self.fields['choices'], self.fields['no_review']
            del self.fields['default_value']
        else:
            if block.type != CustomBlock.InputType.CHOICE:
                del self.fields['choices']
            if block.type != CustomBlock.InputType.NUMERIC:
                del self.fields['numeric_min'], self.fields['numeric_max']
            if block.type == CustomBlock.InputType.TEXT or not block.dependence:
                del self.fields['default_value']
            elif block.type == CustomBlock.InputType.NUMERIC:
                self.fields['default_value'].coerce = lambda val: int(val)
                self.fields['default_value'].msg = 'Enter a whole number.'
            elif block.type == CustomBlock.InputType.BOOLEAN:
                self.fields['default_value'].coerce = \
                    lambda val: {'true': True, 'false': False}[val.lower()]
                self.fields['default_value'].msg = "Enter 'true' or 'false.'"
            elif block.type == CustomBlock.InputType.CHOICE:
                self.fields['default_value'].coerce = \
                    lambda val: { k: k for k in block.choices() }[val]
                self.fields['default_value'].msg = 'Must be one of the choices.'
            
            if block.form.status != Form.Status.DRAFT:
                del self.fields['name']
                if block.type == CustomBlock.InputType.CHOICE:
                    del self.fields['choices']
    
    def clean(self):
        super().clean()
        cleaned_data = self.cleaned_data
        
        if 'default_value' in cleaned_data:
            field = cleaned_data['default_value']
            if not cleaned_data['dependence']:
                cleaned_data['options'].pop('default_value', None)
            elif self.instance.type == CustomBlock.InputType.CHOICE:
                if field and field not in self.instance.choices():
                    cleaned_data['options'].pop('default_value', None)
            elif self.instance.type == CustomBlock.InputType.NUMERIC:
                if field and self.instance.numeric_min() is not None:
                    if field < self.instance.numeric_min():
                        self.add_error('default_value',
                                       "Can't be less than the minimum value")
                elif field and self.instance.numeric_max() is not None:
                    if field > self.instance.numeric_max():
                        self.add_error('default_value',
                                       "Can't be more than the maximum value")
        return cleaned_data


class CollectionBlockAdminForm(FormBlockAdminForm, AdminJSONForm):
    name1 = forms.CharField(required=False, label='text input name 1')
    name2 = forms.CharField(required=False, label='text input name 2')
    name3 = forms.CharField(required=False, label='text input name 3')
    no_review = NegatedBooleanField(label='show in review', required=False)
    button_text = forms.CharField(
        required=False,
        help_text="Label for the collection's 'add' button. "
                  "Defaults to 'add item' or 'add file' or 'add files.'"
    )
    file_types = DynamicArrayField(
        forms.ChoiceField(choices=[ (n, n) for n in FileType.types.keys() ]),
        required=False,
        help_text='Available types are: '+', '.join(FileType.types.keys())+'. '
                  'Leave empty to allow any file type.'
    ) # TODO what if the set of available file types (or stocks) changes?
    max_filesize = forms.IntegerField(
        required=False, help_text='bytes. Applies to any type of file.',
        widget=widgets.AdminIntegerFieldWidget
    )
    autoinit_filename = forms.ChoiceField(
        required=False,
        help_text="If selected, the text field's default will be the file name."
    )
    wide = SingletonChoiceField(
        required=False, label='wide column',
        help_text="If selected, the text field's column will be twice as wide "
                  "as the others."
    )
    choices = DynamicArrayField(
        forms.CharField(max_length=CollectionBlock.FIXED_CHOICE_VAL_MAXLEN),
    )
    
    class Meta:
        exclude = ('form', '_rank', 'align_type')
        static_fields = ('name', 'page', 'fixed', 'name1', 'name2', 'name3',
                         'has_file', 'min_items', 'max_items', 'file_optional',
                         'dependence', 'negate_dependencies')
        json_fields = {'options': ['no_review']}
        dynamic_fields = True
    
    def init_file_fields(self, block):
        admin_fields, total_fields = {}, {}
        self._meta.json_fields['options.file_limits'] = []
        self._meta.json_fields['options.file_processing'] = []
        for name in block.allowed_filetypes() or []:
            filetype = FileType.by_type(name)()
            fields, processing_fields = {}, {}
            for n in filetype.admin_limit_fields():
                fields['min_' + n] = forms.IntegerField()
                fields['max_' + n] = forms.IntegerField()
            for n in filetype.admin_total_fields():
                total_fields['min_' + n] = forms.IntegerField()
                total_fields['max_' + n] = forms.IntegerField()
            for n in filetype.admin_processing_fields():
                processing_fields[n] = forms.IntegerField()
            
            if fields:
                admin_fields[name] = SplitDictField(
                    fields, required=False,
                    widget=SplitDictWidget(fields, two_column=True),
                    label=name+' limits'
                )
                self._meta.json_fields['options.file_limits'].append(name)
            
            if processing_fields:
                proc = name + '_proc'
                admin_fields[proc] = SplitDictField(
                    processing_fields, required=False,
                    widget=SplitDictWidget(processing_fields),
                    label=name+' processing',
                    help_text='File will be processed to conform to the '
                              'limits, if any are given.'
                )
                self._meta.json_fields['options.file_processing'].append(proc)
        
        admin_fields['total'] = SplitDictField(
            total_fields, required=False,
            widget=SplitDictWidget(total_fields, two_column=True),
            label='total limits',
            help_text='Limits that apply to the total value, for all '
                      'files in the collection (when field is applicable).'
        )
        self._meta.json_fields['options.file_limits'].append('total')
        
        return admin_fields

    def __init__(self, *args, **kwargs):
        block, file_limits = None, {}
        if 'instance' in kwargs and kwargs['instance']:
            block = kwargs['instance']
            fields = ['button_text']
            if block.has_file:
                fields += ['file_types', 'max_filesize', 'autoinit_filename']
                self._meta.json_fields['options'] += fields
                
                admin_fields = self.init_file_fields(block)
                kwargs['admin_fields'] = admin_fields
            
            if block.fixed: self._meta.json_fields['options'].append('choices')
            if block.name1 and block.name2:
                self._meta.json_fields['options'].append('wide')
        
        super().__init__(*args, **kwargs)
        
        if not block: del self.fields['no_review'], self.fields['button_text']
        elif block.form.status != Form.Status.DRAFT: del self.fields['name']
        if not block or not block.fixed: del self.fields['choices']

        if block: choices = [ (n, n) for n in block.collection_fields() ]
        if not block or not block.name1 or not block.name2:
            del self.fields['wide']
        else: self.fields['wide'].choices = [(None, '-')] + choices
        
        if not block or not block.has_file:
             for n in ('file_types', 'max_filesize', 'autoinit_filename'):
                del self.fields[n]
        else: self.fields['autoinit_filename'].choices = [(None, '-')] + choices
    
    def clean(self):
        super().clean()
        cleaned_data = self.cleaned_data
        
        if not self.instance.pk:
            if cleaned_data['fixed'] and not cleaned_data['name1']:
                msg = 'A fixed collection must have at least one text input'
                self.add_error('name1', msg)
            return cleaned_data
        
        if 'autoinit_filename' in cleaned_data:
            field = cleaned_data['autoinit_filename']
            if field not in self.instance.collection_fields():
                cleaned_data['options'].pop('autoinit_filename', None)
        return cleaned_data


class SubmissionAdminForm(forms.ModelForm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
        for name, field in self.fields.items():
            if name == '_email': field.required = True
            elif field.required: field.required = False


class SubmissionItemAdminForm(forms.ModelForm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
        for n, field in self.fields.items():
            if isinstance(field, forms.ModelChoiceField): continue
            if n in [ f.name for f in  SubmissionItem._meta.fields ]: continue
            if field.required: field.required = False


class EmailAdminForm(forms.Form):
    name = forms.ChoiceField(choices=(), initial='confirmation')
    subject = forms.CharField(widget=widgets.AdminTextInputWidget)
    content = forms.CharField(
        widget=widgets.AdminTextareaWidget(attrs={'rows': 8, 'cols': 50}),
        help_text="The email's template can optionally be customized, here."
    )
    
    def __init__(self, form=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
        if not form: return
        self.fields['name'].choices = [ (n, n) for n in form.email_names() ]
        self.fields['name'].choices.append(('', '[untitled]'))


class MoveBlocksAdminForm(forms.Form):
    page = forms.ChoiceField(
        choices=(), widget=forms.Select(attrs={'style': 'padding: 4px'})
    )
    
    def __init__(self, max_page, min_page=1, new_page=True, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
        choices = [ (n, f'{n}') for n in range(min_page, max_page + 1) ]
        if new_page: choices.append((max_page + 1, f'{max_page+1} (new)'))
        self.fields['page'].choices = choices



"""
Django settings for Formative project.
"""

import os, environ
from pkg_resources import iter_entry_points
from pathlib import Path

env = environ.Env()

# Build paths inside the project like this: BASE_DIR / 'subdir'.
BASE_DIR = Path(__file__).resolve().parent

environ.Env.read_env(os.path.join(BASE_DIR, '.env'), overwrite=True)
ENV = env('DJANGO_ENV')

SECRET_KEY = env('SECRET_KEY')

DEBUG = env.bool('DEBUG', default=False)

WSGI_APPLICATION = "config.wsgi.application"

ALLOWED_HOSTS = env('ALLOWED_HOSTS', default='').split(',')
csrf_hosts = [ host[0] == '.' and '*' + host or host for host in ALLOWED_HOSTS ]

DEFAULT_SERVER_HOSTNAME = env('SERVER_HOSTNAME')
SERVER_HOSTNAME = env('DJANGO_SERVER_HOSTNAME', default='')
if not SERVER_HOSTNAME: SERVER_HOSTNAME = DEFAULT_SERVER_HOSTNAME

port = env('DJANGO_SERVER_PORT', default=None)
DJANGO_SERVER = SERVER_HOSTNAME + (port and ':' + port or '')

CSRF_TRUSTED_ORIGINS = [ f'https://{DJANGO_SERVER}' ]
if DEBUG:
    CSRF_TRUSTED_ORIGINS = [ f'http://{host}:8000' for host in csrf_hosts ]

CSRF_COOKIE_NAME = 'XSRF-TOKEN'

# Application definition

INSTALLED_APPS = [
    'jazzmin',
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'django.forms',
    'django_better_admin_arrayfield',
    'polymorphic',
    'webpack_loader',
    'widget_tweaks',
    'formative',
]

PLUGINS = []
for entry_point in iter_entry_points(group='formative.plugin', name=None):
    PLUGINS.append(entry_point.module_name)
    INSTALLED_APPS.append(entry_point.module_name)


MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'whitenoise.middleware.WhiteNoiseMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

ROOT_URLCONF = 'urls'

FORM_RENDERER = 'django.forms.renderers.TemplatesSetting'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [os.path.join(BASE_DIR, 'templates')],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]


host = env('POSTGRES_HOST', default='')
if env('DOCKER_ENV', default=''):
    host = env('POSTGRES_DOCKER_HOST')

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.postgresql_psycopg2',
        'NAME': env('POSTGRES_DB'),
        'USER': env('POSTGRES_USER'),
        'PASSWORD': env('POSTGRES_PASSWORD'),
        'HOST': host,
        'PORT': '5432',
    }
}


EMAIL_HOST = env('POSTFIX_HOST', default='localhost')
TECH_EMAIL = env('DJANGO_SU_EMAIL')
CONTACT_EMAIL = env('CONTACT_EMAIL', default=TECH_EMAIL)
SERVER_EMAIL = env('SERVER_EMAIL', default=CONTACT_EMAIL)
ADMINS = [(env('ADMIN_NAME', default=''),
           env('ADMIN_EMAIL', default=TECH_EMAIL))]


prefix = 'django.contrib.auth.password_validation'
AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': prefix + '.UserAttributeSimilarityValidator',
    },
    {
        'NAME': prefix + '.MinimumLengthValidator',
    },
    {
        'NAME': prefix + '.CommonPasswordValidator',
    },
    {
        'NAME': prefix + '.NumericPasswordValidator',
    },
]


LANGUAGE_CODE = 'en-us'

USE_I18N = True

TIME_ZONE = 'UTC'
USE_TZ = True


LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'filters': {
        'require_debug_false': {
            '()': 'django.utils.log.RequireDebugFalse',
        },
        'require_debug_true': {
            '()': 'django.utils.log.RequireDebugTrue',
        },
    },
    'handlers': {
        'console': {
            'level': 'INFO',
            'class': 'logging.StreamHandler',
        },
        'mail_admins': {
            'level': 'ERROR',
            'filters': ['require_debug_false'],
            'class': 'django.utils.log.AdminEmailHandler',
        },
        'mail_debug_admins': {
            'level': 'CRITICAL',
            'filters': ['require_debug_true'],
            'class': 'django.utils.log.AdminEmailHandler',
        }
    },
    'loggers': {
        'django': {
            'handlers': ['console', 'mail_admins', 'mail_debug_admins'],
            'level': 'INFO',
        }
    }
}


STATIC_URL = '/static/'

STATIC_ROOT = os.path.join(os.path.dirname(BASE_DIR), 'static')

STATICFILES_STORAGE = 'whitenoise.storage.CompressedStaticFilesStorage'

MEDIA_ROOT = os.path.join(os.path.dirname(BASE_DIR), 'media')
MEDIA_URL = 'media/'

STATICFILES_DIRS = (
    ("bundles", os.path.join(BASE_DIR, 'assets/bundles')),
#    ("img", os.path.join(BASE_DIR, 'assets/img')),
    os.path.join(BASE_DIR, 'assets/static')
)


WEBPACK_LOADER = {
    'DEFAULT': {
        'CACHE': not DEBUG,
        'BUNDLE_DIR_NAME': 'bundles/',
        'STATS_FILE': os.path.join(os.path.join(BASE_DIR, 'assets/bundles/',
                                                f'webpack-bundle.{ENV}.json')),
        'POLL_INTERVAL': 0.5,
        'TIMEOUT': None,
        'IGNORE': [r".+\.hot-update.js", r".+\.map"]
    }
}


DEFAULT_AUTO_FIELD = 'django.db.models.BigAutoField'


JAZZMIN_SETTINGS = {
    'related_modal_active': True,
    'hide_models': ['formative.collectionblock', 'formative.customblock',
                    'formative.formblock'],
}

JAZZMIN_UI_TWEAKS = {
    'sidebar_nav_legacy_style': True,
    'sidebar_disable_expand': True,
}


X_FRAME_OPTIONS = 'SAMEORIGIN'
SILENCED_SYSTEM_CHECKS = ['security.W019']



# Generated by Django 4.0.2 on 2022-03-03 19:11

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('formative', '0004_alter_form_db_slug_alter_form_slug_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='formblock',
            name='_rank',
            field=models.IntegerField(default=0, null=True, verbose_name=''),
        ),
    ]



from django.core.exceptions import ValidationError
from django.db import models
from localflavor.us import forms as us_forms

from . import CompositeStockWidget


class AddressWidget(CompositeStockWidget):
    TYPE = 'address'
    
    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)
        
        self.template_name = 'address.html'
        self.review_template_name = 'address_review.html'
        
        self.labels = {
            'street_address': 'Street address',
            'city': 'City',
            'state': 'State',
            'postal_code': 'Zip/postal code',
            'country': 'Country'
        }
    
    def fields(self):
        ret = []
        for name in ('street_address', 'city', 'state', 'postal_code',
                      'country'):
            length = 32
            if name == 'street_address': length = 64
            elif name == 'postal_code': length = 16
            
            f = self.field_name(name)
            if name != 'country':
                ret.append((f, models.CharField(max_length=length, blank=True)))
            else:
                ret.append((f, models.CharField(max_length=2, default='US',
                                                choices=COUNTRIES)))
        return ret
    
    def field_required(self, part):
        if not super().field_required(part): return False
        if part in ('street_address', 'city', 'country'): return True
        return False

    def clean(self, data):
        if data['country'] == 'US':
            for name, us_field in (('state', 'USStateField'),
                                   ('postal_code', 'USZipCodeField')):
                if not data[name]:
                    return {name: ValidationError('This field is required.')}
                else:
                    try:
                        val = getattr(us_forms, us_field)().clean(data[name])
                    except ValidationError as e:
                        return { name: e }
                    else:
                        data[name] = val
        return data
    
    def render_choices(self):
        return [('city', 'city, state and country'),
                ('state', 'state and country'), ('country', 'country')]
    
    def render(self, choice, **kwargs):
        vals = [kwargs['country']]
        if choice in ('city', 'state'): vals.insert(0, kwargs['state'])
        if choice == 'city': vals.insert(0, kwargs['city'])
        return ', '.join([ v for v in vals if v ])


COUNTRIES = (
    ('US', 'United States'),
    ('--', '-------------'),
    ('AF', 'Afghanistan'), 
    ('AX', 'Aland Islands'),
    ('AL', 'Albania'), 
    ('DZ', 'Algeria'), 
    ('AS', 'American Samoa'),
    ('AD', 'Andorra'), 
    ('AO', 'Angola'), 
    ('AI', 'Anguilla'), 
    ('AG', 'Antigua and Barbuda'), 
    ('AR', 'Argentina'), 
    ('AM', 'Armenia'), 
    ('AW', 'Aruba'), 
    ('AU', 'Australia'), 
    ('AT', 'Austria'), 
    ('AZ', 'Azerbaijan'), 
    ('BS', 'Bahamas (the)'), 
    ('BH', 'Bahrain'), 
    ('BD', 'Bangladesh'), 
    ('BB', 'Barbados'), 
    ('BY', 'Belarus'), 
    ('BE', 'Belgium'), 
    ('BZ', 'Belize'), 
    ('BJ', 'Benin'), 
    ('BM', 'Bermuda'), 
    ('BT', 'Bhutan'), 
    ('BO', 'Bolivia (Plurinational State of)'), 
    ('BQ', 'Bonaire, Sint Eustatius and Saba'),
    ('BA', 'Bosnia and Herzegovina'), 
    ('BW', 'Botswana'), 
    ('BR', 'Brazil'), 
    ('BN', 'Brunei Darussalam'), 
    ('BG', 'Bulgaria'), 
    ('BF', 'Burkina Faso'), 
    ('BI', 'Burundi'), 
    ('CV', 'Cabo Verde'), 
    ('KH', 'Cambodia'), 
    ('CM', 'Cameroon'), 
    ('CA', 'Canada'), 
    ('KY', 'Cayman Islands (the)'), 
    ('CF', 'Central African Republic (the)'), 
    ('TD', 'Chad'), 
    ('CL', 'Chile'), 
    ('CN', 'China'), 
    ('CX', 'Christmas Island'),
    ('CC', 'Cocos (Keeling) Islands (the)'),
    ('CO', 'Colombia'), 
    ('KM', 'Comoros'), 
    ('CG', 'Congo (the)'), 
    ('CD', 'Congo (the Democratic Republic of the)'),
    ('CK', 'Cook Islands (the)'),
    ('CR', 'Costa Rica'), 
    ('CI', 'Cote d\'Ivoire'), 
    ('HR', 'Croatia'), 
    ('CU', 'Cuba'), 
    ('CW', 'Curacao'),
    ('CY', 'Cyprus'), 
    ('CZ', 'Czechia'),
    ('DK', 'Denmark'), 
    ('DJ', 'Djibouti'), 
    ('DM', 'Dominica'), 
    ('DO', 'Dominican Republic'), 
    ('EC', 'Ecuador'), 
    ('EG', 'Egypt'), 
    ('SV', 'El Salvador'), 
    ('GQ', 'Equatorial Guinea'), 
    ('ER', 'Eritrea'), 
    ('EE', 'Estonia'), 
    ('SZ', 'Eswatini'), 
    ('ET', 'Ethiopia'), 
    ('FK', 'Falkland Islands (the)'), 
    ('FO', 'Faroe Islands (the)'), 
    ('FJ', 'Fiji'), 
    ('FI', 'Finland'), 
    ('FR', 'France'), 
    ('GF', 'French Guiana'), 
    ('PF', 'French Polynesia'), 
    ('GA', 'Gabon'), 
    ('GM', 'Gambia (the)'), 
    ('GE', 'Georgia'), 
    ('DE', 'Germany'), 
    ('GH', 'Ghana'), 
    ('GI', 'Gibraltar'),
    ('GR', 'Greece'), 
    ('GL', 'Greenland'), 
    ('GD', 'Grenada'), 
    ('GP', 'Guadeloupe'), 
    ('GT', 'Guatemala'), 
    ('GG', 'Guernsey'),
    ('GN', 'Guinea'), 
    ('GW', 'Guinea-Bissau'), 
    ('GY', 'Guyana'), 
    ('HT', 'Haiti'), 
    ('VA', 'Holy See'), 
    ('HN', 'Honduras'), 
    ('HK', 'Hong Kong'), 
    ('HU', 'Hungary'), 
    ('IS', 'Iceland'), 
    ('IN', 'India'), 
    ('ID', 'Indonesia'), 
    ('IR', 'Iran (Islamic Republic of)'),
    ('IQ', 'Iraq'), 
    ('IE', 'Ireland'), 
    ('IM', 'Isle of Man'),
    ('IL', 'Israel'), 
    ('IT', 'Italy'), 
    ('JM', 'Jamaica'), 
    ('JP', 'Japan'), 
    ('JE', 'Jersey'),
    ('JO', 'Jordan'), 
    ('KZ', 'Kazakhstan'), 
    ('KE', 'Kenya'), 
    ('KI', 'Kiribati'), 
    ('KP', 'Korea (DPRK)'),
    ('KR', 'Korea (the Republic of)'),
    ('KW', 'Kuwait'), 
    ('KG', 'Kyrgyzstan'), 
    ('LA', 'Lao People\'s Democratic Republic (the)'),
    ('LV', 'Latvia'), 
    ('LB', 'Lebanon'), 
    ('LS', 'Lesotho'), 
    ('LR', 'Liberia'), 
    ('LY', 'Libya'),
    ('LI', 'Liechtenstein'), 
    ('LT', 'Lithuania'), 
    ('LU', 'Luxembourg'), 
    ('MO', 'Macao'), 
    ('MK', 'Macedonia'),
    ('MG', 'Madagascar'), 
    ('MW', 'Malawi'), 
    ('MY', 'Malaysia'), 
    ('MV', 'Maldives'), 
    ('ML', 'Mali'), 
    ('MT', 'Malta'), 
    ('MH', 'Marshall Islands (the)'),
    ('MQ', 'Martinique'), 
    ('MR', 'Mauritania'), 
    ('MU', 'Mauritius'), 
    ('YT', 'Mayotte'),
    ('MX', 'Mexico'), 
    ('FM', 'Micronesia (Federated States of)'),
    ('MD', 'Moldova (the Republic of)'), 
    ('MC', 'Monaco'), 
    ('MN', 'Mongolia'), 
    ('ME', 'Montenegro'), 
    ('MS', 'Montserrat'), 
    ('MA', 'Morocco'), 
    ('MZ', 'Mozambique'), 
    ('MM', 'Myanmar'), 
    ('NA', 'Namibia'), 
    ('NR', 'Nauru'), 
    ('NP', 'Nepal'), 
    ('NL', 'Netherlands (the)'), 
    ('NC', 'New Caledonia'), 
    ('NZ', 'New Zealand'), 
    ('NI', 'Nicaragua'), 
    ('NE', 'Niger'), 
    ('NG', 'Nigeria'), 
    ('NU', 'Niue'), 
    ('NF', 'Norfolk Island'),
    ('MP', 'Northern Mariana Islands (the)'),
    ('NO', 'Norway'), 
    ('OM', 'Oman'), 
    ('PK', 'Pakistan'), 
    ('PW', 'Palau'), 
    ('PS', 'Palestine, State of'), 
    ('PA', 'Panama'), 
    ('PG', 'Papua New Guinea'), 
    ('PY', 'Paraguay'), 
    ('PE', 'Peru'), 
    ('PH', 'Philippines'), 
    ('PN', 'Pitcairn'), 
    ('PL', 'Poland'), 
    ('PT', 'Portugal'), 
    ('PR', 'Puerto Rico'), 
    ('QA', 'Qatar'), 
    ('RE', 'Reunion'), 
    ('RO', 'Romania'), 
    ('RU', 'Russian Federation (the)'), 
    ('RW', 'Rwanda'), 
    ('BL', 'Saint Barthelemy'),
    ('SH', 'Saint Helena, Ascension and TdC'), 
    ('KN', 'Saint Kitts and Nevis'), 
    ('LC', 'Saint Lucia'), 
    ('MF', 'Saint Martin (French part)'),
    ('PM', 'Saint Pierre and Miquelon'), 
    ('VC', 'Saint Vincent and the Grenadines'), 
    ('WS', 'Samoa'), 
    ('SM', 'San Marino'), 
    ('ST', 'Sao Tome and Principe'), 
    ('SA', 'Saudi Arabia'), 
    ('SN', 'Senegal'), 
    ('RS', 'Serbia'), 
    ('SC', 'Seychelles'), 
    ('SL', 'Sierra Leone'), 
    ('SG', 'Singapore'), 
    ('SX', 'Sint Maarten (Dutch part)'),
    ('SK', 'Slovakia'), 
    ('SI', 'Slovenia'), 
    ('SB', 'Solomon Islands'), 
    ('SO', 'Somalia'), 
    ('ZA', 'South Africa'), 
    ('SS', 'South Sudan'),
    ('ES', 'Spain'), 
    ('LK', 'Sri Lanka'), 
    ('SD', 'Sudan'), 
    ('SR', 'Suriname'), 
    ('SJ', 'Svalband and Jan Mayen'),
    ('SE', 'Sweden'), 
    ('CH', 'Switzerland'), 
    ('SY', 'Syrian Arab Republic'),
    ('TW', 'Taiwan'),
    ('TJ', 'Tajikistan'), 
    ('TZ', 'Tanzania, United Republic of'),
    ('TH', 'Thailand'), 
    ('TL', 'Timor-Leste'), 
    ('TG', 'Togo'), 
    ('TK', 'Tokelau'),
    ('TO', 'Tonga'), 
    ('TT', 'Trinidad and Tobago'), 
    ('TN', 'Tunisia'), 
    ('TR', 'Turkey'), 
    ('TM', 'Turkmenistan'), 
    ('TC', 'Turks and Caicos Islands (the)'), 
    ('TV', 'Tuvalu'), 
    ('UG', 'Uganda'), 
    ('UA', 'Ukraine'), 
    ('AE', 'United Arab Emirates'), 
    ('GB', 'United Kingdom'), 
    ('UY', 'Uruguay'), 
    ('UZ', 'Uzbekistan'), 
    ('VU', 'Vanuatu'), 
    ('VE', 'Venezuela (Bolivarian Republic of)'), 
    ('VN', 'Viet Nam'), 
    ('VG', 'Virgin Islands (British)'), 
    ('VI', 'Virgin Islands (U.S.)'),
    ('WF', 'Wallis and Futuna'), 
    ('EH', 'Western Sahara'), 
    ('YE', 'Yemen'), 
    ('ZM', 'Zambia'), 
    ('ZW', 'Zimbabwe'), 
)




__all__ = ["OpenForm", "SubmissionForm", "ItemFileForm", "ItemsForm",
           "ItemsFormSet", "AdminJSONForm", "ProgramAdminForm", "FormAdminForm",
           "StockBlockAdminForm", "CustomBlockAdminForm",
           "CollectionBlockAdminForm", "SubmissionAdminForm",
           "SubmissionItemAdminForm", "EmailAdminForm"]

from .formative import *
from .admin import *



from django.db.models import Q, Exists, OuterRef
from django.db.models.signals import pre_save, post_save, \
    pre_delete, post_delete
from django.apps import apps
from django.dispatch import Signal, dispatcher, receiver
from django.utils.text import capfirst

from .models import Form, FormBlock, CustomBlock, CollectionBlock, FormLabel
from .stock import EmailWidget
from .utils import any_name_field


@receiver(post_save, sender=Form)
def form_post_save(sender, instance, created, raw, **kwargs):
    if raw or not created: return

    form = instance
    if form.validation_type == Form.Validation.EMAIL:
        b = FormBlock(form=form, page=0, name='email',
                      options=EmailWidget.default_options())
        b.save()

@receiver(pre_delete, sender=Form)
def form_pre_delete(sender, instance, **kwargs):
    form = instance

    if form.status != Form.Status.DRAFT:
        form.unpublish()

@receiver(pre_save, sender=CustomBlock)
def customblock_pre_save(sender, instance, raw, **kwargs):
    if raw: return
    
    if instance.pk:
        orig = CustomBlock.objects.get(pk=instance.pk)
        instance._old_name = orig.name

def default_text(text):
    return capfirst(text.replace('_', ' '))

@receiver(post_save, sender=CustomBlock)
def customblock_post_save(sender, instance, created, raw, **kwargs):
    if raw: return

    block, new = instance, created
    if not block.page: return # no autocreated labels for autocreated fields

    # set up default labels for the block
    style = FormLabel.LabelStyle.WIDGET
    if block.type == CustomBlock.InputType.TEXT:
        style = block.form.default_text_label_style()
        if block.num_lines > 1: style = FormLabel.LabelStyle.VERTICAL
    elif block.type == CustomBlock.InputType.CHOICE:
        style = FormLabel.LabelStyle.VERTICAL
    elif block.type == CustomBlock.InputType.NUMERIC:
        style = FormLabel.LabelStyle.HORIZONTAL

    text = default_text(block.name)
    if style != FormLabel.LabelStyle.WIDGET: text += ':' # TODO: i18n
    
    l = None
    if new:
        l = FormLabel(form=block.form, path=block.name, style=style, text=text)
    elif block.name != block._old_name:
        try:
            l = FormLabel.objects.get(form=block.form, path=block._old_name,
                                      style=style)
            l.path = block.name
        except FormLabel.DoesNotExist: pass
    if l: l.save()
    
    if block.type != CustomBlock.InputType.CHOICE: return
    paths = []
    for c in 'choices' in block.options and block.choices() or []:
        text = default_text(c)
        new_path = '.'.join((block.name, c))
        if not new and block.name != block._old_name:
            path = '.'.join((block._old_name, c))
        else: path=new_path
        
        l = None
        if not new:
            try:
                l = FormLabel.objects.get(form=block.form, path=path,
                                          style=FormLabel.LabelStyle.WIDGET)
                l.path = new_path
            except FormLabel.DoesNotExist: pass
        
        if not l:
            l = FormLabel(form=block.form, path=new_path,
                          style=FormLabel.LabelStyle.WIDGET, text=text)
        
        l.save()
        paths.append(l.path)
    
    delete_sub_labels(block.form, block.name, paths)
    if not new and block._old_name != block.name:
        delete_sub_labels(block.form, block._old_name)

@receiver(pre_save, sender=FormBlock)
def formblock_pre_save(sender, instance, raw, **kwargs):
    if raw: return
    
    if instance.pk:
        orig = FormBlock.objects.get(pk=instance.pk)
        instance._old_name = orig.name

@receiver(post_save, sender=FormBlock)
def formblock_post_save(sender, instance, created, raw, **kwargs):
    if raw: return

    block, stock, new = instance, instance.stock, created

    # get stock block's FormLabels and save
    paths = []
    for new_path, (style, text) in stock.widget_labels().items():
        if not new and block.name != block._old_name:
            if '.' in new_path:
                widget = new_path[new_path.index('.')+1:]
                path = '.'.join((block._old_name, widget))
            else: path = block._old_name
        else: path = new_path
        
        l = None
        if not new:
            try:
                l = FormLabel.objects.get(form=block.form, path=path,
                                          style=style)
                l.path = new_path
            except FormLabel.DoesNotExist: pass
        
        if not l:
            l = FormLabel(form=block.form, path=new_path, style=style,
                          text=text)
        l.save()
        paths.append(l.path)
    
    delete_sub_labels(block.form, block.name, paths)
    if not new and block._old_name != block.name:
        delete_sub_labels(block.form, block._old_name)

@receiver(post_save, sender=CollectionBlock)
def collectionblock_post_save(sender, instance, created, raw, **kwargs):
    block = instance
    if raw: return
    
    fields = block.collection_fields()
    
    if created:
        existing = block.form.custom_blocks().filter(name__in=fields, page=0)
        existing_names = existing.values_list('name', flat=True)
        for name in fields:
            if name in existing_names: continue
        
            item_field = CustomBlock.text_create(form=block.form, name=name,
                                                 page=0)
            item_field.save()
    
    refs = CollectionBlock.objects.filter(any_name_field(_=OuterRef('name')),
                                          form=block.form)
    block.form.collections().filter(~Exists(refs), page=0, _rank__gt=1).delete()
    
    if not created: return
    
    text, style = default_text(block.name) + ':', FormLabel.LabelStyle.VERTICAL
    l = FormLabel.objects.get_or_create(form=block.form, path=block.name,
                                        defaults={'style': style, 'text': text})
    
    if block.fixed: return # no implementation yet for fixed + collection
    style = FormLabel.LabelStyle.WIDGET
    for name in block.collection_fields():
        text, path = default_text(name), '.'.join((block.name, name))
        l = FormLabel.objects.get_or_create(form=block.form, path=path,
                                            defaults={'style': style,
                                                      'text': text})

def delete_sub_labels(form, name, exclude=[]):
    sl = form.labels.filter(path__startswith=name+'.').exclude(path__in=exclude)
    sl.delete()

def delete_block_labels(form, name):
    form.labels.filter(Q(path=name) | Q(path__startswith=name+'.')).delete()

@receiver(post_delete, sender=CustomBlock)
def customblock_post_delete(sender, instance, **kwargs):
    delete_block_labels(instance.form, instance.name)

@receiver(post_delete, sender=FormBlock)
def formblock_post_delete(sender, instance, **kwargs):
    delete_block_labels(instance.form, instance.name)

@receiver(post_delete, sender=CollectionBlock)
def collectionblock_post_delete(sender, instance, **kwargs):
    block = instance
    
    name_with_id = f'{block.name}{block.id}'
    block.form.labels.filter(Q(path=name_with_id+'_') |
                             Q(path__startswith=name_with_id+'.',
                               path__endswith='_')).delete()
    
    blocks = block.form.collections(name=block.name)
    if not blocks:
        delete_block_labels(block.form, block.name)
        return
    
    for name in block.collection_fields():
        if not blocks.filter(any_name_field(_=name)):
            block.form.labels.filter(path='.'.join((block.name, name))).delete()


app_cache = {}

def populate_app_cache():
    global app_cache
    apps.check_apps_ready()
    for config in apps.app_configs.values(): app_cache[config.name] = config

    
class FormPluginSignal(Signal):
    def _is_active(self, sender, receiver):
        searchpath = receiver.__module__
        app = None
        while True:
            app = app_cache.get(searchpath)
            if '.' not in searchpath or app: break
            searchpath, _ = searchpath.rsplit('.', 1)
        
        return sender and app and app.name in sender.get_plugins()
    
    def send(self, sender, **kwargs):
        if sender and not isinstance(sender, Form):
            raise ValueError('Signal sender needs to be a form.')
        
        responses = []
        if not self.receivers: return responses
        if self.sender_receivers_cache.get(sender) is dispatcher.NO_RECEIVERS:
            return responses
        
        if not app_cache: populate_app_cache()
        
        for receiver in self._live_receivers(sender):
            if self._is_active(sender, receiver):
                response = receiver(signal=self, sender=sender, **kwargs)
                responses.append((receiver, response))
        return responses


register_program_settings = Signal()

register_form_settings = FormPluginSignal()

submission_review_pre = FormPluginSignal()

submission_review_post = FormPluginSignal()

submission_submit_control = FormPluginSignal()

submission_thanks = FormPluginSignal()

submission_handle_submit = FormPluginSignal()



# Generated by Django 4.0.3 on 2022-03-22 01:38

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('formative', '0005_alter_formblock__rank'),
    ]

    operations = [
        migrations.AlterField(
            model_name='formblock',
            name='_rank',
            field=models.IntegerField(default=0, verbose_name=''),
            preserve_default=False,
        ),
    ]



from django.forms import BooleanField
from django.utils.text import capfirst

__all__ = ["StockWidget", "EmailWidget", "NameWidget", "AddressWidget",
           "PhoneNumberWidget", "URLWidget", "ChoiceSetWidget", "DateWidget"]


class StockWidget:
    types = {}
    composite = False
    
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if cls.TYPE: cls.types[cls.TYPE] = cls
    
    @classmethod
    def by_type(cls, type):
        return cls.types[type]
    
    @classmethod
    def default_options(cls):
        return { 'type': cls.TYPE }
    
    def __init__(self, name, required=False, **kwargs):
        self.name = name
        self.required = required
    
    def widget_names(self):
        return (self.name,)

    def field_name(self, *args):
        return '_' + self.name
    
    def field_names(self):
        return (self.field_name(),)

    def field_required(self, widget=None):
        return self.required
    
    def field_validators(self, widget=None):
        return []

    def default_label(self):
        from ..signals import default_text
        
        return default_text(self.name)
    
    def get_widget(self, field_name):
        return None # methods will be called w/ widget=None since all one field

    def widget_labels(self):
        from ..models import FormLabel
        
        return {
            self.name: (FormLabel.LabelStyle.WIDGET, self.default_label())
        }
    
    # (need to clarify 'widget' in this context) form_widget is Django's
    def form_widget(self, name):
        return None # use default
    
    def conditional_value(self, **kwargs):
        # default is to return a boolean that's True if we got some input
        return bool([v for v in kwargs.values() if v])
    
    def clean(self, data):
        return data
    
    def admin_fields(self):
        return {'required': BooleanField(required=False)}
    
    def admin_clean(self, data):
        return data
    
    def admin_published_readonly(self):
        return {}


class CompositeStockWidget(StockWidget):
    TYPE = None
    composite = True

    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)
        
        # dict mapping path names to default widget labels
        self.labels = {}

    def widget_names(self): return self.labels.keys()
    
    def field_names(self):
        return tuple(self.field_name(f) for f in self.widget_names())
    
    def field_name(self, field):
        # name has an initial underscore so that non-stock fields can't conflict
        return f'_{self.name}_{field}'

    def get_widget(self, field_name):
        return field_name[1:][len(self.name)+1:]
    
    def widget_labels(self):
        from ..models import FormLabel
        
        labels = {}
        for name, label in self.labels.items():
            LabelStyle = FormLabel.LabelStyle
            if name is None: labels[self.name] = (LabelStyle.VERTICAL, label)
            else: labels[f'{self.name}.{name}'] = (LabelStyle.WIDGET, label)
        return labels
    
    def render_choices(self):
        return [('', '-')]
    
    # method to render a composite widget's values into a single string
    def render(self, choice, **kwargs):
        return ''


from .email import EmailWidget
from .name import NameWidget
from .address import AddressWidget
from .phone import PhoneNumberWidget
from .url import URLWidget
from .choiceset import ChoiceSetWidget
from .date import DateWidget



from django.utils.translation import gettext_lazy as _
from pikepdf import Pdf

from . import FileType


class DocumentFile(FileType):
    TYPE = 'document'
    EXTENSIONS = ('pdf',)
    
    def meta(self, path):
        ret = super().meta(path)
        
        try:
            with Pdf.open(path) as pdf:
                pages = len(pdf.pages)
            ret.update(pages=pages)
            return ret
        
        except:
            msg = _('Error occurred reading the PDF file.')
            self.logger.critical(msg, exc_info=True)
            return {'error': msg}
    
    def admin_limit_fields(self):
        return ('pages',)
    
    def admin_total_fields(self):
        return ('pages',)



from django import template
from django.utils.safestring import mark_safe

import importlib


register = template.Library()


@register.simple_tag
def form_signal(form, signame, **kwargs):
    sigmodule, signame = signame.rsplit('.', 1)
    sigmodule = importlib.import_module(sigmodule)
    signal = getattr(sigmodule, signame)
    
    html_result = []
    for _, response in signal.send(form, **kwargs):
        if response: html_result.append(response)
    return mark_safe(''.join(html_result))



from django.core.exceptions import ValidationError
from django.db import models
from django.forms import RadioSelect, CheckboxInput, CharField, BooleanField, \
    IntegerField
from django.utils.text import capfirst
from django.utils.translation import gettext_lazy as _
from django_better_admin_arrayfield.forms.fields import DynamicArrayField

from . import CompositeStockWidget


class ChoiceSetWidget(CompositeStockWidget):
    TYPE = 'choiceset'
    
    def __init__(self, name, choices=[], single=False, text_input=None,
                 text_input_maxlength=64, **kwargs):
        from ..signals import default_text
        
        super().__init__(name, **kwargs)
        
        self.template_name = 'choiceset.html'
        self.review_template_name = 'choiceset_review.html'
        
        self.single = single
        self.choices = choices
        self.labels = { name: default_text(name) for name in choices }
        if text_input: self.labels[text_input] = capfirst(text_input)
        self.text_input = text_input
        self.text_input_maxlength = text_input_maxlength
    
    def widget_names(self): # override for the ordering
        if self.single: names = [self.name]
        else: names = self.choices
        if self.text_input: return names + [self.text_input]
        return names
    
    def widget_labels(self):
        from ..models import FormLabel
        
        labels = super().widget_labels()
        labels[self.name] = (FormLabel.LabelStyle.VERTICAL,
                             self.default_label() + ':')
        return labels
    
    def fields(self):
        cls = models.BooleanField
        args = {'null': True}
        
        if not self.single:
            fields = [ (self.field_name(field), cls(**args))
                       for field in self.choices ]
        else:
            ml = max(len(name) for name in self.choices)
            choices = [ (k, self.labels[k]) for k in self.choices ]
            fields = [(self.field_name(self.name),
                      models.CharField(blank=True, max_length=ml,
                                       choices=choices, **args))]
        
        if self.text_input:
            ml = self.text_input_maxlength
            fields.append((self.field_name(self.text_input),
                           models.CharField(blank=True, max_length=ml, **args)))
        return fields
    
    def field_required(self, part):
        req = super().field_required(part)
        # TODO: single choice, required == True needs special handling w/ clean
        if self.single and req and part != self.text_input: return True
        return False
    
    def form_widget(self, name):
        if name != self.text_input:
            if self.single: return RadioSelect
            return CheckboxInput
        return super().form_widget(name)
    
    def admin_fields(self):
        from ..models import CustomBlock
        
        choices = DynamicArrayField(
            CharField(max_length=CustomBlock.CHOICE_VAL_MAXLEN)
        )
        single = BooleanField(label='single selection', required=False)
        text_input = CharField(label='text input ID', required=False,
                               max_length=30)
        max_length = IntegerField(label='max text characters', required=False,
                                  min_value=1, max_value=1000)
        
        f = super().admin_fields()
        f.update({'choices': choices, 'single': single,
                  'text_input': text_input, 'text_input_maxlength': max_length})
        return f
    
    def admin_clean(self, data):
        if 'text_input' in data and data['text_input']:
            if not data['text_input_maxlength']:
                err = ValidationError('required if there is a text input',
                                      code='required')
                return {'text_input_maxlength': err}
        return data
    
    def admin_published_readonly(self):
        return {'choices': 'choices', 'text_input': 'text input ID'}
    
    def render_choices(self):
        return [('with', 'with unlisted'), ('without', 'with [unlisted]'),
                ('exclude', 'without unlisted')]
    
    def render(self, choice, **kwargs):
        vals = []
        if self.single and kwargs[self.name]: vals.append(kwargs[self.name])
        else:
            for field in self.choices:
                if kwargs[field]: vals.append(field) # TODO: relabel?
        if choice != 'exclude' and self.text_input:
            if kwargs[self.text_input]:
                if choice == 'with': vals.append(kwargs[self.text_input])
                else: vals.append('[unlisted]')
            
        return ', '.join(vals)



from django import template
from django.conf import settings
import os

from .. import __version__
from ..utils import thumbnail_path, human_readable_filesize

register = template.Library()

@register.simple_tag
def get_formative_version():
    return __version__

@register.simple_tag
def file_thumbnail(file):
    path = thumbnail_path(file.path, ext='jpg')
    if os.path.isfile(path): return thumbnail_path(file.url, ext='jpg')
    path = thumbnail_path(file.path)
    if os.path.isfile(path): return thumbnail_path(file.url)
    
    return None

register.filter('human_readable', human_readable_filesize)

@register.filter
def underscore(obj, name):
    attr = getattr(obj, '_' + name)
    if callable(attr): return attr()
    return attr



from django.core import validators
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _
import re

from .utils import get_file_extension, human_readable_filesize


class WordValidator(validators.BaseValidator):
    pattern = re.compile(r'\w+')
    
    def clean(self, x):
        return len(self.pattern.findall(x))


class MinWordsValidator(WordValidator):
    message = _('Ensure this text has at least %(limit_value)d words' +
                ' (it has %(show_value)d).')
    code = 'min_words'

    def compare(self, a, b):
        return a < b


class MaxWordsValidator(WordValidator):
    message = _('Ensure this text has at most %(limit_value)d words' +
                ' (it has %(show_value)d).')
    code = 'max_words'

    def compare(self, a, b):
        return a > b


class FileExtensionValidator(validators.FileExtensionValidator):
    MAX_EXTENSION_LENGTH = 10
    message = 'Allowed file extensions are: %(allowed_extensions)s.'
    
    def __call__(self, value):
        if self.allowed_extensions is None:
            # validate that there is some extension
            extension = get_file_extension(value)
            if not extension or len(extension) > self.MAX_EXTENSION_LENGTH:
                message = _('File name must end with an extension.')
                raise ValidationError(message, code='no_extension',
                                      params={'extension': extension,
                                              'value': value})
            return
        
        class NameAsFile:
            pass
        f = NameAsFile()
        f.name = value
        
        super().__call__(f)


class FileSizeValidator(validators.MaxValueValidator):
    message = _('Maximum file size is %(val_with_unit)s.')
    
    def __call__(self, value):
        params = {'val_with_unit': human_readable_filesize(self.limit_value)}
        if self.compare(value, self.limit_value):
            raise ValidationError(self.message, code=self.code, params=params)


def validate_identifier(name):
    validators.RegexValidator(
        r'^(_|[0-9])|_(i|u)$',
        'Identifiers cannot start with a number or "_", or end with "_i" '
        'or "_u".',
        inverse_match=True
    )(name.lower().replace('-', ''))

def validate_program_identifier(name):
    validators.RegexValidator(
        r'^(formative|auth|django|silk)$',
        'The identifier cannot be "formative," "auth," "django," or "silk."',
        inverse_match=True
    )(name.lower().replace('-', ''))
    
    validate_identifier(name)

def validate_form_identifier(name):
    validators.RegexValidator(
        r'^(submissionrecord|program|form|formlabel|formdependency'
        r'|formblock|customblock|collectionblock)$',
        'The identifier cannot be "form," "program," or any of the other '
        'Formative internal names.',
        inverse_match=True
    )(name.lower().replace('-', ''))
    
    validate_identifier(name)

def validate_formblock_identifier(name):
    validate_identifier(name)



from django.utils.translation import gettext_lazy as _
from PIL import Image, ImageFile, ImageOps
import os

from ..utils import thumbnail_path
from . import FileType

ImageFile.LOAD_TRUNCATED_IMAGES = True


class ImageFile(FileType):
    TYPE = 'image'
    EXTENSIONS = ('jpg', 'jpeg', 'gif', 'png')
    
    def meta(self, path):
        ret = super().meta(path)
        
        try:
            with Image.open(path) as img:
                width, height = img.size
                icc = img.info.get('icc_profile', '')
                profile = ''
                if icc:
                    profile = ''.join(chr(x) for x in icc[4:8] if x) + ' '
                    profile += ''.join(chr(x) for x in icc[48:56] if x)
            
            ret.update(width=width, height=height,
                       icc_profile=profile,
                       megapixels=width*height/1000000)
            # TODO: check that img.format matches the extension
            return ret
        except:
            msg = _('Error occurred reading the image file.')
            self.logger.critical(msg, exc_info=True)
            return {'error': msg}
    
    def process(self, file, meta, max_width=None, max_height=None, **kwargs):
        try:
            if 'icc_profile' not in meta or not meta['icc_profile']:
                # some browsers mess up w/ untagged images - should assume sRGB
                pass # TODO needs more research
#                with Image.open(file.path) as img:
#                    srgb = ImageCms.createProfile('sRGB')
#                    img = ImageCms.profileToProfile(orig, srgb).tobytes()
#                    img.save(file.path, img.format, icc_profile=srgb_profile)
#                    meta['icc_profile'] = ' '
            
            if not max_width and not max_height: return meta
            if meta['width'] <= max_width and meta['height'] <= max_height:
                return meta
            
            with Image.open(file.path) as img:
                img.load()
                profile = img.info.get('icc_profile', '')
                img.thumbnail((max_width, max_height), Image.ANTIALIAS)
                new_img = ImageOps.exif_transpose(img)
                new_width, new_height = new_img.size
                new_img.save(file.path, img.format, icc_profile=profile)
                new_size = os.path.getsize(file.path)
            
            meta['width'], meta['height'] = new_width, new_height
            meta['megapixels'] = new_width * new_height / 1000000
            msg = _('Image was resized to %(width)sx%(height)s.')
            meta['message'] = msg % {'width': new_width, 'height': new_height}
            meta['update_filesize'] = new_size
            return meta
        
        except:
            msg = _('Error occurred processing the image file.')
            self.logger.critical(msg, exc_info=True)
            return {'error': msg}
    
    def submitted(self, items):
        for item in items:
            file = item._file
            try:
                with Image.open(file.path) as img:
                    profile = img.info.get('icc_profile', '')
                    if img.mode != 'RGB': img = img.convert('RGB')
                    img.thumbnail((120, 120), Image.ANTIALIAS)
                    new_img = ImageOps.exif_transpose(img)
                    # TODO: need an alternative name thing
                    outpath = thumbnail_path(file.path)
                    if not os.path.isfile(outpath):
                        new_img.save(outpath, img.format, icc_profile=profile)
            except:
                self.logger.critical('Error generating thumbnail.',
                                     exc_info=True)
    
    def admin_limit_fields(self):
        return ('width', 'height', 'megapixels')
    
    def admin_total_fields(self):
        return ('megapixels',)
    
    def admin_processing_fields(self):
        return ('max_width', 'max_height')



from django.core.exceptions import ValidationError
from django.db import models
from django.contrib.admin.widgets import AdminRadioSelect
from django.forms import ChoiceField
from django.utils.translation import gettext_lazy as _

from . import CompositeStockWidget


class NameWidget(CompositeStockWidget):
    TYPE = 'name'
    
    class Parts:
        ALL = 'all'
        LAST = 'last'
        ANY = 'any'
    
    def __init__(self, name, required_part=Parts.ALL, **kwargs):
        super().__init__(name, **kwargs)
        
        self.template_name = 'name.html'
        self.review_template_name = 'name_review.html'
        
        self.labels = {
            'firstname': 'First name',
            'lastname': 'Last name'
        }
        self.required_part = required_part
    
    def widget_names(self): return ('firstname', 'lastname')
    
    def fields(self):
        cls = models.CharField
        args = {'max_length': 32, 'blank': True}
        
        return [ (self.field_name(field), cls(**args))
                 for field in self.widget_names() ]
    
    def field_required(self, part):
        if not super().field_required(part): return False
        if self.required_part == self.Parts.ANY: return False
        if self.required_part == self.Parts.LAST and part == 'lastname':
            return True
        if self.required_part == self.Parts.ALL: return True
        return False
    
    def clean(self, data):
        if self.required_part == self.Parts.ANY:
            if all(not data[w] for w in self.widget_names()):
                msg = _('One of the name fields must be provided.')
                return {None: ValidationError(msg)}
        return data
    
    def admin_fields(self):
        required_part = ChoiceField(required=False, widget=AdminRadioSelect,
            choices=[(self.Parts.LAST, 'last name'),
                     (self.Parts.ANY, 'any part'),
                     (self.Parts.ALL, 'all parts')]
        )
        
        f = super().admin_fields()
        f['required_part'] = required_part
        return f
    
    def render_choices(self):
        return [('full', 'Full name')]
    
    def render(self, choice, **kwargs):
        return f"{kwargs['firstname']} {kwargs['lastname']}"



import os
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = []
    
    def generate_superuser(apps, schema_editor):
        from django.contrib.auth.models import User
        
        DJANGO_SU_NAME = os.environ.get('DJANGO_SU_NAME')
        DJANGO_SU_EMAIL = os.environ.get('DJANGO_SU_EMAIL')
        DJANGO_SU_PASSWORD = os.environ.get('DJANGO_SU_PASSWORD')

        superuser = User.objects.create_superuser(
            username=DJANGO_SU_NAME,
            email=DJANGO_SU_EMAIL,
            password=DJANGO_SU_PASSWORD)
        superuser.save()

    operations = [
        migrations.RunPython(generate_superuser),
    ]



from django.apps import apps
from django.urls import include, path
from django.views.defaults import page_not_found, server_error
from django.http import Http404
from django.contrib.auth.views import LoginView

from formative import admin

import importlib.util


plugin_patterns = []
for app in apps.get_app_configs():
    if hasattr(app, 'FormativePluginMeta'):
        if importlib.util.find_spec(app.name + '.urls'):
            urlmod = importlib.import_module(app.name + '.urls')
            plugin_patterns.append(path('', include((urlmod.urlpatterns,
                                                     app.label))))

urlpatterns = [
    path('admin/', admin.site.urls),
    path('accounts/login/', LoginView.as_view(template_name='admin/login.html',
                                              next_page='/')),
    path('accounts/', include('django.contrib.auth.urls')),
    path('', include('formative.urls')),
    path('', include((plugin_patterns, 'plugins'))),
#    path('404/', page_not_found, {'exception': Http404()}),
#    path('500/', server_error),
]



# Generated by Django 4.0.2 on 2022-02-03 19:39

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('formative', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Form',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=64)),
                ('slug', models.SlugField(allow_unicode=True, editable=False, max_length=64)),
                ('db_slug', models.SlugField(allow_unicode=True, editable=False, max_length=64)),
                ('status', models.CharField(choices=[('draft', 'unpublished'), ('disabled', 'submissions disabled'), ('enabled', 'published/enabled'), ('completed', 'completed')], default='draft', max_length=16)),
                ('options', models.JSONField(blank=True, default=dict)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(default=django.utils.timezone.now, editable=False)),
                ('completed', models.DateTimeField(blank=True, editable=False, null=True)),
                ('validation_type', models.CharField(choices=[('email', 'email address')], default='email', editable=False, max_length=16)),
            ],
        ),
        migrations.CreateModel(
            name='FormBlock',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('_rank', models.IntegerField(default=0, editable=False)),
                ('name', models.SlugField(allow_unicode=True, max_length=32, verbose_name='identifier')),
                ('options', models.JSONField(blank=True, default=dict)),
                ('page', models.PositiveIntegerField(default=1)),
                ('negate_dependencies', models.BooleanField(default=False, verbose_name='negate dependency')),
                ('dependence', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='dependents', related_query_name='dependent', to='formative.formblock')),
                ('form', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='blocks', related_query_name='block', to='formative.form')),
                ('polymorphic_ctype', models.ForeignKey(editable=False, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='polymorphic_%(app_label)s.%(class)s_set+', to='contenttypes.contenttype')),
            ],
            options={
                'ordering': ['form', 'page', '_rank'],
                'abstract': False,
                'base_manager_name': 'objects',
            },
        ),
        migrations.CreateModel(
            name='Program',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=32)),
                ('slug', models.SlugField(allow_unicode=True, editable=False, max_length=32, unique=True)),
                ('db_slug', models.SlugField(allow_unicode=True, editable=False, max_length=32, unique=True)),
                ('description', models.CharField(blank=True, max_length=250)),
                ('options', models.JSONField(blank=True, default=dict)),
                ('hidden', models.BooleanField(default=False)),
                ('created', models.DateTimeField(auto_now_add=True)),
            ],
            options={
                'ordering': ['created'],
            },
        ),
        migrations.CreateModel(
            name='CollectionBlock',
            fields=[
                ('block', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='formative.formblock')),
                ('fixed', models.BooleanField(default=False)),
                ('min_items', models.PositiveIntegerField(blank=True, null=True)),
                ('max_items', models.PositiveIntegerField(blank=True, null=True)),
                ('has_file', models.BooleanField(default=False)),
                ('file_optional', models.BooleanField(default=False)),
                ('name1', models.CharField(blank=True, default='', max_length=32)),
                ('name2', models.CharField(blank=True, default='', max_length=32)),
                ('name3', models.CharField(blank=True, default='', max_length=32)),
                ('align_type', models.CharField(choices=[('horizontal', 'horizontal'), ('vertical', 'vertical')], default='horizontal', max_length=16)),
            ],
            options={
                'db_table': 'formative_formcollectionblock',
            },
            bases=('formative.formblock',),
        ),
        migrations.CreateModel(
            name='CustomBlock',
            fields=[
                ('block', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='formative.formblock')),
                ('type', models.CharField(choices=[('text', 'text'), ('num', 'numeric'), ('choice', 'multiple choice'), ('bool', 'true/false choice')], default='text', max_length=16)),
                ('required', models.BooleanField(default=False)),
                ('num_lines', models.PositiveIntegerField(default=1)),
                ('min_chars', models.PositiveIntegerField(blank=True, null=True)),
                ('max_chars', models.PositiveIntegerField(blank=True, null=True)),
                ('min_words', models.PositiveIntegerField(blank=True, null=True)),
                ('max_words', models.PositiveIntegerField(blank=True, null=True)),
            ],
            options={
                'db_table': 'formative_formcustomblock',
            },
            bases=('formative.formblock',),
        ),
        migrations.CreateModel(
            name='FormLabel',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('path', models.CharField(max_length=128)),
                ('text', models.CharField(max_length=1000)),
                ('style', models.CharField(choices=[('widget', 'widget label'), ('vertical', 'vertical label'), ('horizontal', 'horizontal label')], default='widget', max_length=16)),
                ('form', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='labels', related_query_name='label', to='formative.form')),
            ],
        ),
        migrations.CreateModel(
            name='FormDependency',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('value', models.CharField(blank=True, max_length=64)),
                ('block', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='dependencies', related_query_name='dependency', to='formative.formblock')),
            ],
            options={
                'verbose_name': 'dependency value',
                'verbose_name_plural': 'dependency values',
            },
        ),
        migrations.AddField(
            model_name='form',
            name='program',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='forms', related_query_name='form', to='formative.program'),
        ),
        migrations.AddConstraint(
            model_name='formlabel',
            constraint=models.UniqueConstraint(fields=('form', 'path', 'style'), name='unique_path_style'),
        ),
        migrations.AddConstraint(
            model_name='formdependency',
            constraint=models.UniqueConstraint(fields=('block', 'value'), name='unique_blockval'),
        ),
        migrations.AddConstraint(
            model_name='formblock',
            constraint=models.UniqueConstraint(fields=('form', 'page', '_rank'), name='unique_rank'),
        ),
        migrations.AddConstraint(
            model_name='form',
            constraint=models.UniqueConstraint(fields=('program', 'slug'), name='unique_slug'),
        ),
        migrations.AddConstraint(
            model_name='form',
            constraint=models.UniqueConstraint(fields=('program', 'db_slug'), name='unique_db_slug'),
        ),
    ]



# Generated by Django 4.0.2 on 2022-02-11 23:33

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('formative', '0002_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='SubmissionRecord',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('form', models.SlugField(allow_unicode=True, max_length=64)),
                ('submission', models.UUIDField(editable=False)),
                ('type', models.CharField(max_length=32)),
                ('recorded', models.DateTimeField(auto_now=True, verbose_name='recorded at')),
                ('text', models.TextField(blank=True)),
                ('number', models.PositiveBigIntegerField(blank=True, null=True)),
                ('deleted', models.BooleanField(default=False)),
                ('program', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='formative.program')),
            ],
        ),
        migrations.AddConstraint(
            model_name='submissionrecord',
            constraint=models.UniqueConstraint(fields=('submission', 'type'), name='unique_submission_record_type'),
        ),
    ]



# Generated by Django 4.0.2 on 2022-02-18 18:56

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('formative', '0003_submissionrecord_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='form',
            name='db_slug',
            field=models.SlugField(allow_unicode=True, editable=False, max_length=30),
        ),
        migrations.AlterField(
            model_name='form',
            name='slug',
            field=models.SlugField(allow_unicode=True, max_length=30, verbose_name='identifier'),
        ),
        migrations.AlterField(
            model_name='program',
            name='db_slug',
            field=models.SlugField(allow_unicode=True, editable=False, max_length=30, unique=True),
        ),
        migrations.AlterField(
            model_name='program',
            name='name',
            field=models.CharField(max_length=64),
        ),
        migrations.AlterField(
            model_name='program',
            name='slug',
            field=models.SlugField(allow_unicode=True, max_length=30, unique=True, verbose_name='identifier'),
        ),
    ]



from django.db import models
from django.utils.html import mark_safe

from . import StockWidget

class URLWidget(StockWidget):
    TYPE = 'url'
    
    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)
        
        self.template_name = 'url.html'
        self.review_template_name = 'review.html'
    
    def fields(self):
        field = models.URLField(blank=True)
        
        return [(self.field_name(), field)]
    
    def clean(self, data):
        if '://' not in data[:8]:
            return 'http://' + data
        return data
    
    def render(self, choice, **kwargs):
        url = kwargs[self.name]
        if not url: return ''
        return mark_safe(f'<a target="_blank" href="{url}">{url}</a>')
        



from django.core.exceptions import ValidationError
from django.db import transaction, models
from django.db.models import Max, Q, F, Case, When, Subquery, \
    ExpressionWrapper, IntegerField


class UnderscoredRankedModel(models.Model):
    class Meta:
        abstract = True
    
    _rank = models.IntegerField(verbose_name='')
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
        self._initial_rank = None
        if '_rank' not in self.get_deferred_fields():
            self._initial_rank = self._rank

    def save(self, *args, **kwargs):
        if not self.pk or not self._rank: # self._rank == 0 for pass-thru
            if self._rank:
                raise ValidationError('New RankedModel instance already'
                                      'ranked - this is not supported')
            elif self._rank == 0: # we're just passing thru to get a lock
                super().save(*args, **kwargs)
                return
            
            # newly created instance: it goes at the end
            with transaction.atomic():
                # insert with zero - we're using this as the table lock
                self._rank = 0
                super().save(*args, **kwargs)
                
                group = self._rank_group()
                query = group.aggregate(max_rank=Max('_rank'))
                if query['max_rank'] is not None:
                    self._rank = query['max_rank'] + 1
                else: self._rank = 1
                
                super().save(*args, **kwargs)
                self._initial_rank = self._rank
                return
        
        n, positive = self._rank - self._initial_rank, True
        if n < 0: n, positive = -n, False
        elif not n: # fast-path for when there's no apparent change:
            self._rank = F('_rank') # if rank has in fact changed, leave it be
            super().save(*args, **kwargs)
            self._rank = self._initial_rank # no refresh; caller is responsible
            return
        
        group = self._rank_group()
        
        # start from where we _really_ are, not where we thought we were;
        # otherwise we're not actually locking the correct rows
        pos = Subquery(group.filter(pk=self.pk).values('_rank')[:1])
        if positive:
            section = group.annotate(pos=pos).filter(_rank__gte=F('pos'),
                                                     _rank__lte=F('pos')+n)
        else:
            section = group.annotate(pos=pos).filter(_rank__lte=F('pos'),
                                                     _rank__gte=F('pos')-n)
        
        with transaction.atomic():
            order = positive and '_rank' or '-_rank'
            query = section.order_by(order).select_for_update()
            count, last_ranked = 0, None
            for count, last_ranked in enumerate(query): pass # rows are locked
            
            # we could have hit the end early:
            if last_ranked: new_rank = last_ranked._rank 
            else: new_rank = self._rank
            
            if not count:
                # nothing to do; we just had a misapprehension
                self._rank = new_rank
                super().save(*args, **kwargs)
                self._initial_rank = self._rank
                return
            
            if positive: # this moving + direction means that the others move -
                section = group.filter(_rank__gt=new_rank-count,
                                       _rank__lte=new_rank)
                # use the negatives as a temporary space, to avoid key conflicts
                section.update(_rank=-F('_rank'))
                increment = -1
                section = group.filter(_rank__gte=-new_rank,
                                       _rank__lt=-new_rank+count)
            else:
                # they move out of the way in the other direction
                section = group.filter(_rank__gte=new_rank,
                                       _rank__lt=new_rank+count)
                section.update(_rank=-F('_rank'))
                increment = 1
                section = group.filter(_rank__gt=-new_rank-count,
                                       _rank__lte=-new_rank)
            
            self._rank = new_rank
            super().save(*args, **kwargs)
            self._initial_rank = self._rank
            
            # make the others positive again, with the increment applied
            section.update(_rank=-F('_rank')+increment)
    
    @transaction.atomic
    def delete(self, *args, **kwargs):
        self.refresh_from_db() # less concerned about performance for this one
        rank = self._rank
        self._rank = 0
        if rank:
            self.save(update_fields=['_rank']) # get the table lock
            group = self._rank_group()
            group.filter(_rank__gt=rank).update(_rank=-F('_rank'))
            group.filter(_rank__lt=-rank).update(_rank=-F('_rank')-1)
        
        super().delete(*args, **kwargs)
    
    def _rank_group(self):
        if hasattr(self, 'rank_group') and callable(self.rank_group):
            return self.rank_group()
        
        return self.__class__.objects.all()


class RankedModel(UnderscoredRankedModel):
    class Meta:
        abstract = True
    
    @property
    def rank(self):
        return self._rank
    
    @rank.setter
    def rank(self, value):
        self._rank = value
    
    def rank_group(self):
        return self.__class__.objects.all()



from django.core.validators import RegexValidator
from django.db import models
from django.utils.translation import gettext_lazy as _

from . import StockWidget


class PhoneNumberWidget(StockWidget):
    TYPE = 'phone'
    
    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)
        
        self.template_name = 'phone.html'
        self.review_template_name = 'review.html'
        
    def fields(self):
        field = models.CharField(max_length=32, blank=True)
        return [(self.field_name(), field)]

    def field_validators(self, widget=None):
        return [RegexValidator(r'^[0-9 .()/+-]*$',
                               message=_('Invalid phone number format.'))]



from django.db import models
from django.core.exceptions import ValidationError
from django.utils.text import slugify


class AutoSlugModel(models.Model):
    class Meta:
        abstract = True
    
    def save(self, *args, **kwargs):
        self.db_slug = self.slug.lower().replace('-', '')
        
        try:
            super().save(*args, **kwargs)
        except ValidationError as e:
            msg = _('Name (with hyphens removed) must be unique.')
            raise ValidationError(msg) from e
